import numpy as np
import networkx as nx


def __knn(k, x):
    from sklearn.neighbors import NearestNeighbors

    neigh = NearestNeighbors(k + 1, metric="euclidean", n_jobs=-1).fit(x)
    k_neighbors = neigh.kneighbors(
        x,
        k + 1,
    )
    distance = np.array(k_neighbors[0][:, 1:])
    indices = np.array(k_neighbors[1][:, 1:])
    return distance, indices


def __w_matrix(distance, indices, ks, a=10):
    n = len(distance)

    weight_matrix = np.zeros([n, n])

    sigma2 = (a / n / ks) * np.linalg.norm(distance) ** 2

    if ks == 1:
        for i in range(n):
            j = indices[i][0]
            weight_matrix[i][j] = np.exp(-1 * (distance[i, 0] ** 2) / sigma2)
    else:
        for i in range(n):
            for k, j in enumerate(indices[i]):
                weight_matrix[i][j] = np.exp(-1 * (distance[i, k] ** 2) / sigma2)

    return weight_matrix, sigma2


def __k0graph(similarity):
    x = np.arange(len(similarity))
    y = [np.argmax(row) for row in similarity]

    vc = []

    for i in range(len(x)):
        x_index, y_index = -1, -1
        for k in range(len(vc)):
            if y[i] in vc[k]:
                y_index = k
            if x[i] in vc[k]:
                x_index = k

        if x_index == y_index and x_index != -1 and y_index != -1:
            continue
        elif x_index < 0 and y_index < 0:
            vc.append([x[i], y[i]])
        elif x_index >= 0 > y_index:
            vc[x_index].append(y[i])
        elif x_index < 0 <= y_index:
            vc[y_index].append(x[i])
        else:
            vc[x_index].extend(vc[y_index])
            del vc[y_index]

    return vc


def __get_affinity_matrix(vc, w):
    nc = len(vc)

    affinity = np.zeros([nc, nc])

    for i in range(nc):
        for j in range(i + 1, nc):
            ij = np.ix_(vc[i], vc[j])
            ji = np.ix_(vc[j], vc[i])

            w_ij, w_ji = w[ij], w[ji]
            ci, cj = len(vc[i]), len(vc[j])

            ones_i = np.ones((ci, 1))
            ones_j = np.ones((cj, 1))
            affinity[i][j] = (1 / ci ** 2) * np.transpose(ones_i).dot(w_ij).dot(
                w_ji
            ).dot(ones_i) + (1 / cj ** 2) * np.transpose(ones_j).dot(w_ji).dot(
                w_ij
            ).dot(
                ones_j
            )
            affinity[j][i] = affinity[i][j]
    return affinity


def __get_affinity_btw_cluster(c1, c2, w):
    ij = np.ix_(c1, c2)
    ji = np.ix_(c2, c1)

    w_ij, w_ji = w[ij], w[ji]
    ci, cj = len(c1), len(c2)

    ones_i = np.ones((ci, 1))
    ones_j = np.ones((cj, 1))
    affinity = (1 / ci ** 2) * np.transpose(ones_i).dot(w_ij).dot(w_ji).dot(ones_i) + (
        1 / cj ** 2
    ) * np.transpose(ones_j).dot(w_ji).dot(w_ij).dot(ones_j)
    return affinity[0, 0]


def __get_neighbor(vc, kc, w):
    ns, as_ = [], []
    A = __get_affinity_matrix(vc, w)

    for i in range(len(A)):
        as_.append([x for x in sorted(list(A[i]))[-1 * kc :] if x > 0])
        n = len(as_[i])
        if n == 0:
            ns.append([])
        else:
            ns.append(A[i].argsort()[-1 * n :].tolist())

    return ns, as_


def preprocess(data, ks, a):
    """
    From data to graph.
    __knn and __w_matrix only use for preprocess
    """
    distance, indices = __knn(ks, data)
    # convert distance to similarity
    similarity, _ = __w_matrix(distance, indices, ks, a)
    g = nx.from_numpy_matrix(similarity, create_using=nx.DiGraph)
    return g


def Agdl(g, target_cluster_num, kc):
    similarity = nx.to_numpy_matrix(g)
    # Using k0grpha to initilize cluster

    cluster = __k0graph(similarity)

    neighbor_set, affinity_set = __get_neighbor(cluster, kc, similarity)
    current_cluster_num = len(cluster)

    while current_cluster_num > target_cluster_num:

        max_affinity = 0
        max_index1 = 0
        max_index2 = 0
        for i in range(len(neighbor_set)):
            if len(neighbor_set[i]) == 0:
                continue
            aff = max(affinity_set[i])
            if aff > max_affinity:
                j = int(neighbor_set[i][affinity_set[i].index(aff)])
                max_affinity = aff

                if i < j:
                    max_index1 = i
                    max_index2 = j
                else:
                    max_index1 = j
                    max_index2 = i

        if max_index1 == max_index2:
            break

        # merge two cluster
        cluster[max_index1].extend(cluster[max_index2])
        cluster[max_index2] = []

        if max_index2 in neighbor_set[max_index1]:
            p = neighbor_set[max_index1].index(max_index2)
            del neighbor_set[max_index1][p]
        if max_index1 in neighbor_set[max_index2]:
            p = neighbor_set[max_index2].index(max_index1)
            del neighbor_set[max_index2][p]

        for i in range(len(neighbor_set)):
            if i == max_index1 or i == max_index2:
                continue

            if max_index1 in neighbor_set[i]:
                aff_update = __get_affinity_btw_cluster(
                    cluster[i], cluster[max_index1], similarity
                )

                p = neighbor_set[i].index(max_index1)
                affinity_set[i][p] = aff_update  # fix the affinity values

            if max_index2 in neighbor_set[i]:

                p = neighbor_set[i].index(max_index2)
                del neighbor_set[i][p]
                del affinity_set[i][p]

                if max_index1 not in neighbor_set[i]:
                    aff_update = __get_affinity_btw_cluster(
                        cluster[i], cluster[max_index1], similarity
                    )
                    neighbor_set[i].append(max_index1)
                    affinity_set[i].append(aff_update)

        neighbor_set[max_index1].extend(neighbor_set[max_index2])
        neighbor_set[max_index1] = list(set(neighbor_set[max_index1]))

        affinity_set[max_index1] = []

        neighbor_set[max_index2] = []
        affinity_set[max_index2] = []

        # Fine the Kc-nearest clusters for Cab

        for i in range(len(neighbor_set[max_index1])):
            target_index = neighbor_set[max_index1][i]
            new_affinity = __get_affinity_btw_cluster(
                cluster[target_index], cluster[max_index1], similarity
            )
            affinity_set[max_index1].append(new_affinity)

        if len(affinity_set[max_index1]) > kc:
            index = np.argsort(affinity_set[max_index1])
            new_neighbor = []
            new_affinity = []
            for j in range(kc):
                new_neighbor.append(neighbor_set[max_index1][index[-1 * j]])
                new_affinity.append(affinity_set[max_index1][index[-1 * j]])

            neighbor_set[max_index1] = new_neighbor
            affinity_set[max_index1] = new_affinity

        current_cluster_num = current_cluster_num - 1

    reduced_cluster = []
    for i in range(len(cluster)):
        if len(cluster[i]) != 0:
            reduced_cluster.append(cluster[i])

    return reduced_cluster



#######################################################################################################################
#               PercoMCV code source, implemented by the www.abil.ac.cd team composed by Elie Mayogha,                #
#                       Selain Kasereka, Nathanael Kasoro, Ho Tuong Vinh and Joel Kinganga                            #
#                                                                                                                     #
#       We invite contributor to reuse our code source and cite our paper. We would like to be contacted when         #
#       this code is used, this way will allow us to know the evolution of our proposed algorithm. Injoy              #
#                                   Contact us: contact@abil.ac.cd - University of Kinshasa                           #
#######################################################################################################################

from collections import defaultdict
import networkx as nx

# First step
# computation of k-clique percolation algorithm
# with k = 4


def __k_clique_communities(g, cliques=None):
    if cliques is None:
        cliques = nx.find_cliques(g)
    cliques = [frozenset(c) for c in cliques if len(c) >= 4]

    # First index which nodes are in which cliques
    membership_dict = defaultdict(list)
    for clique in cliques:
        for node in clique:
            membership_dict[node].append(clique)

    # For each clique, see which adjacent cliques percolate
    perc_graph = nx.Graph()
    perc_graph.add_nodes_from(cliques)
    for clique in cliques:
        for adj_clique in _get_adjacent_cliques(clique, membership_dict):
            if len(clique.intersection(adj_clique)) >= 3:
                perc_graph.add_edge(clique, adj_clique)

    # Connected components of clique graph with perc edges
    # are the percolated cliques
    for component in nx.connected_components(perc_graph):
        yield frozenset.union(*component)


def _get_adjacent_cliques(clique, membership_dict):
    adjacent_cliques = set()
    for n in clique:
        for adj_clique in membership_dict[n]:
            if clique != adj_clique:
                adjacent_cliques.add(adj_clique)
    return adjacent_cliques


def percoMVC(g):
    # Zachary's Karate club example
    c = list(__k_clique_communities(g))

    m2 = set()
    coms = []
    for com in c:
        coms_list = list(com)
        coms += [coms_list]
        m1 = set(coms_list)
        m2 = m2 | m1

    t = []
    p = 1
    while p <= len(g.nodes):
        t.append(p)
        p += 1
    t = set(t)
    nodn_classes = t - m2

    # Second step
    # Trying to classify unclassified nodes7

    nodn_classes = sorted(nodn_classes)

    for Com in range(len(coms)):
        if len(coms[Com]) > 3:  # Check si la communauté à plus de 3 noeud

            sub = g.subgraph(coms[Com])

            # Calcul de la centralité de vecteur propre
            centrality = nx.eigenvector_centrality(sub)
            vercteur_pr = sorted(
                (round((centrality[node]), 2), node) for node in centrality
            )
            for vect in range(len(vercteur_pr)):
                centralitiness = (
                    vercteur_pr[vect][0] / vercteur_pr[len(vercteur_pr) - 1][0]
                )
                if centralitiness >= 0.99:  # check if the node is 99% central
                    neud_central = vercteur_pr[vect][1]
                    for nod in range(len(nodn_classes)):
                        if g.has_edge(nodn_classes[nod], neud_central):
                            coms[Com] += [nodn_classes[nod]]

    return coms



from cdlib.classes.node_clustering import NodeClustering
from cdlib.classes.edge_clustering import EdgeClustering
from cdlib.classes.fuzzy_node_clustering import FuzzyNodeClustering
from cdlib.classes.attr_node_clustering import AttrNodeClustering
from cdlib.classes.bipartite_node_clustering import BiNodeClustering
from cdlib.classes.temporal_clustering import TemporalClustering
from cdlib.classes.named_clustering import NamedClustering



import numpy as np
from scipy import sparse
from collections import deque
from collections import defaultdict
import networkx as nx


class MultiCom(object):
    # https://github.com/ahollocou/multicom
    #
    # Hollocou, Alexandre and Bonald, Thomas and Lelarge, Marc
    # "Multiple Local Community Detection"
    # International Symposium on Computer Performance, Modeling, Measurements and Evaluation 2017

    def __init__(self, g):
        self.g = g

    def __load_graph(self):
        """
        Load an undirected and unweighted graph from an edge-list file.
        :param edgelist_filename: string or unicode
            Path to the edge-list file.
            Id of nodes are assumed to be non-negative integers.
        :param delimiter: str, default '\t'
        :param comment: str, default '#'
        :return: Compressed Sparse Row Matrix
            Adjacency matrix of the graph
        """
        edge_df = nx.to_pandas_edgelist(self.g)
        edge_list = edge_df.values
        n_nodes = int(np.max(edge_list) + 1)
        adj_matrix = sparse.coo_matrix(
            (np.ones(edge_list.shape[0]), (edge_list[:, 0], edge_list[:, 1])),
            shape=tuple([n_nodes, n_nodes]),
            dtype=edge_list.dtype,
        )
        adj_matrix = adj_matrix.tocsr()
        adj_matrix = adj_matrix + adj_matrix.T
        return adj_matrix

    def __convert_adj_matrix(self, adj_matrix):
        """
        Convert an adjacency matrix to the Compressed Sparse Row type.
        :param adj_matrix: An adjacency matrix.
        :return: Compressed Sparse Row Matrix
            Adjacency matrix with the expected type.
        """
        if type(adj_matrix) == sparse.csr_matrix:
            return adj_matrix
        elif type(adj_matrix) == np.ndarray:
            return sparse.csr_matrix(adj_matrix)
        else:
            raise TypeError(
                "The argument should be a Numpy Array or a Compressed Sparse Row Matrix."
            )

    def __approximate_ppr(self, adj_matrix, seed_set, alpha=0.85, epsilon=1e-5):
        """
        Compute the approximate Personalized PageRank (PPR) from a set set of seed node.
        This function implements the push method introduced by Andersen et al.
        in "Local graph partitioning using pagerank vectors", FOCS 2006.
        :param adj_matrix: compressed sparse row matrix or numpy array
            Adjacency matrix of the graph
        :param seed_set: list or set of int
            Set of seed nodes.
        :param alpha: float, default 0.85
            1 - alpha corresponds to the probability for the random walk to restarts from the seed set.
        :param epsilon: float, default 1e-3
            Precision parameter for the approximation
        :return: numpy 1D array
            Vector containing the approximate PPR for each node of the graph.
        """
        adj_matrix = self.__convert_adj_matrix(adj_matrix)
        degree = np.array(np.sum(adj_matrix, axis=0))[0]
        n_nodes = adj_matrix.shape[0]

        prob = np.zeros(n_nodes)
        res = np.zeros(n_nodes)
        res[list(seed_set)] = 1.0 / len(seed_set)

        next_nodes = deque(seed_set)

        while len(next_nodes) > 0:
            node = next_nodes.pop()
            push_val = res[node] - 0.5 * epsilon * degree[node]
            res[node] = 0.5 * epsilon * degree[node]
            prob[node] += (1.0 - alpha) * push_val
            put_val = alpha * push_val
            for neighbor in adj_matrix[node].indices:
                old_res = res[neighbor]
                res[neighbor] += put_val * adj_matrix[node, neighbor] / degree[node]
                threshold = epsilon * degree[neighbor]
                if res[neighbor] >= threshold > old_res:
                    next_nodes.appendleft(neighbor)
        return prob

    def __conductance_sweep_cut(self, adj_matrix, score, window=10):
        """
        Return the sweep cut for conductance based on a given score.
        During the sweep process, we detect a local minimum of conductance using a given window.
        The sweep process is described by Andersen et al. in
        "Communities from seed sets", 2006.
        :param adj_matrix: compressed sparse row matrix or numpy array
            Adjacency matrix of the graph.
        :param score: numpy vector
            Score used to order the nodes in the sweep process.
        :param window: int, default 10
            Window parameter used for the detection of a local minimum of conductance.
        :return: set of int
             Set of nodes corresponding to the sweep cut.
        """
        adj_matrix = self.__convert_adj_matrix(adj_matrix)
        n_nodes = adj_matrix.shape[0]
        degree = np.array(np.sum(adj_matrix, axis=0))[0]
        total_volume = np.sum(degree)
        sorted_nodes = [node for node in range(n_nodes) if score[node] > 0]
        sorted_nodes = sorted(sorted_nodes, key=lambda node: score[node], reverse=True)
        sweep_set = set()
        volume = 0.0
        cut = 0.0
        best_conductance = 1.0
        best_sweep_set = {sorted_nodes[0]}
        inc_count = 0
        for node in sorted_nodes:
            volume += degree[node]
            for neighbor in adj_matrix[node].indices:
                if neighbor in sweep_set:
                    cut -= 1
                else:
                    cut += 1
            sweep_set.add(node)

            if volume == total_volume:
                break
            conductance = cut / min(volume, total_volume - volume)
            if conductance < best_conductance:
                best_conductance = conductance
                # Make a copy of the set
                best_sweep_set = set(sweep_set)
                inc_count = 0
            else:
                inc_count += 1
                if inc_count >= window:
                    break
        return best_sweep_set

    def execute(self, seed_node, clustering=None, n_steps=5, explored_ratio=0.8):
        """
        Algorithm for multiple local algorithms detection from a seed node.
        It implements the algorithm presented by Hollocou, Bonald and Lelarge in
        "Multiple Local Community Detection".
        :param g: networkx graph
        :param seed_node: int
            Id of the seed node around which we want to detect communities.
        :param clustering: Scikit-Learn Cluster Estimator
            Algorithm used to cluster nodes in the local embedding space.
            Example: sklearn.cluster.DBSCAN()
        :param n_steps: int, default 5
            Parameter used to control the number of detected communities.
        :param explored_ratio: float, default 0.8
            Parameter used to control the number of new seeds at each step.
        :return:
        seeds: list of int
            Seeds used to detect communities around the initial seed (including this original seed).
        communities: list of set
            Communities detected around the seed node.
        """
        seeds = dict()
        scores = dict()
        communities = list()
        explored = set()

        if clustering is None:
            from sklearn.cluster import DBSCAN

            clustering = DBSCAN()

        adj_matrix = self.__load_graph()
        adj_matrix = self.__convert_adj_matrix(adj_matrix)
        n_nodes = adj_matrix.shape[0]
        degree = np.array(np.sum(adj_matrix, axis=0))[0]

        new_seeds = [seed_node]
        step = -1
        n_iter = 0

        while step < n_steps and len(new_seeds) > 0:
            n_iter += 1

            for new_seed in new_seeds:
                step += 1
                seeds[step] = new_seed
                scores[step] = self.__approximate_ppr(adj_matrix, [seeds[step]])
                community = self.__conductance_sweep_cut(adj_matrix, scores[step])
                communities.append(community)
                # We add the algorithms to the explored nodes
                explored |= set(community)

            new_seeds = list()

            # Clustering of the nodes in the space (scores[seed1], scores[seed2],...,)
            embedding = np.zeros((n_nodes, step + 1))
            for i in range(step + 1):
                embedding[:, i] = scores[i][:]
            indices = np.where(np.sum(embedding, axis=1))[0]
            y = clustering.fit_predict(embedding[indices, :])
            clusters = defaultdict(set)
            for i in range(y.shape[0]):
                if y[i] != -1:
                    clusters[y[i]].add(indices[i])

            # Pick new seeds in unexplored clusters
            for c in range(len(clusters)):
                cluster_size = 0
                cluster_explored = 0
                for node in clusters[c]:
                    cluster_size += 1
                    if node in explored:
                        cluster_explored += 1
                if float(cluster_explored) / float(cluster_size) < explored_ratio:
                    candidates = list(set(clusters[c]) - explored)
                    candidate_degrees = np.array([degree[node] for node in candidates])
                    new_seeds.append(candidates[np.argmax(candidate_degrees)])

        return list(communities)

    @staticmethod
    def __get_node_membership(communities):
        """
        Get the algorithms membership for each node given a list of communities.
        :param communities: list of list of int
            List of communities.
        :return: membership: dict (defaultdict) of set of int
            Dictionary such that, for each node,
            membership[node] is the set of algorithms ids to which the node belongs.
        """
        membership = defaultdict(set)
        for i, community in enumerate(communities):
            for node in community:
                membership[node].add(i)
        return membership



# True Price
from itertools import combinations
from collections import defaultdict


def i_pca(g, weights=None, t_in=0.5):
    if weights is None:
        return __ipca(g, t_in)
    else:
        return __ipca_weighted(g, weights, t_in)


def __ipca(g, t_in=0.5):
    res = []
    data = defaultdict(set)  # node id => neighboring node ids

    # read in graph
    for a, b in g.edges():
        data[a].add(b)
        data[b].add(a)
    unvisited = []
    weights = defaultdict(int)
    for a, b in combinations(data, 2):
        if b in data[a]:
            shared = len(data[a] & data[b])
            weights[a] += shared
            weights[b] += shared

        unvisited = set(data)
    num_clusters = 0

    seed_nodes = sorted(data, key=lambda k: (weights[k], len(data[k])), reverse=True)

    for seed in seed_nodes:  # get highest degree node
        if seed in unvisited:
            cluster = {v for v in data[seed]}  # seed and random neighbor

            while True:
                # rank neighbors by the number of edges between the node and cluster nodes
                frontier = sorted(
                    (len(data[p] & cluster), p)
                    for p in set.union(*((data[n] - cluster) for n in cluster))
                )

                # do this until IN_vk < T_IN, SP <= 2 is met, or no frontier nodes left
                found = False
                while frontier and not found:
                    m_vk, p = frontier.pop()
                    if m_vk < t_in * len(cluster):
                        break
                    c_2neighbors = data[p] & cluster
                    c_2neighbors.update(*(data[c] & cluster for c in c_2neighbors))
                    if cluster == c_2neighbors:
                        found = True
                        break

                # otherwise, add the node to the cluster
                if found:
                    cluster.add(p)
                    continue

                break

            unvisited -= cluster
            res.append(list(cluster))

            num_clusters += 1

            if not unvisited:
                break
    return res


def __ipca_weighted(g, wlabel, t_in):
    data = defaultdict(set)  # node id => neighboring node ids
    global_edges, degrees = defaultdict(dict), defaultdict(float)
    res = []

    # read in graph
    for a, b, dt in g.edges(data=True):
        data[a].add(b)
        data[b].add(a)
        w = dt[wlabel]
        global_edges[a][b], global_edges[b][a] = w, w
        degrees[a] += w
        degrees[b] += w

    weights = defaultdict(int)
    for a, b in combinations(data, 2):
        if b in data[a]:
            shared = len(data[a] & data[b])
            weights[a] += shared
            weights[b] += shared

    unvisited = set(data)
    num_clusters = 0

    # DIFFERENT: weighted degrees
    seed_nodes = sorted(data, key=lambda k: (weights[k], degrees[k]), reverse=True)

    for seed in seed_nodes:  # get highest weight/degree node
        if seed in unvisited:
            cluster = {seed}

            while True:
                # DIFFERENT: rank neighbors by the sum of global edge weights
                # between the node and cluster nodes
                frontier = sorted(
                    (sum(global_edges[p][c] for c in data[p] & cluster), p)
                    for p in set.union(*((data[n] - cluster) for n in cluster))
                )

                # do this until IN_vK < T_IN, SP <= 2 is met, or no frontier nodes left
                found = False
                while frontier and not found:
                    m_vk, p = frontier.pop()
                    if m_vk < t_in * len(cluster):
                        break
                    c_2neighbors = data[p] & cluster
                    c_2neighbors.update(*(data[c] & cluster for c in c_2neighbors))
                    if cluster == c_2neighbors:
                        found = True
                        break

                if not found:
                    break

                cluster.add(p)  # otherwise, add the node to the cluster

            unvisited -= cluster
            if len(cluster) > 1:
                res.append(list(cluster))

                num_clusters += 1

            if unvisited:
                pass
            else:
                break
    return res



# Author: True Price <jtprice@cs.unc.edu>

from collections import defaultdict


def m_code(g, weights=None, weight_threshold=0.2):
    weight_threshold = 1 - weight_threshold
    if weights is None:
        return __mcode_unweighted(g, weight_threshold)
    else:
        return __mcode_weighted(g, weights, weight_threshold)


def __mcode_unweighted(g, weight_threshold):
    res = []
    edges = defaultdict(set)  # node id => neighboring node ids

    # read in graph

    for a, b in g.edges():
        edges[a].add(b)
        edges[b].add(a)

    # Stage 1: Vertex Weighting
    weights = dict((v, 1.0) for v in edges)

    for i, v in enumerate(edges):
        neighborhood = {v} | edges[v]
        # see if larger k-cores exist
        if len(neighborhood) > 2:
            k = 1  # highest valid k-core
            while neighborhood:
                k_core = neighborhood.copy()
                invalid_nodes = True
                while invalid_nodes and neighborhood:
                    invalid_nodes = set(
                        n for n in neighborhood if len(edges[n] & neighborhood) <= k
                    )
                    neighborhood -= invalid_nodes
                k += 1  # on exit, k will be one greater than we want
            # vertex weight = k-core number * density of k-core
            weights[v] = (k - 1) * (
                sum(len(edges[n] & k_core) for n in k_core) / (2.0 * len(k_core) ** 2)
            )

        # if node has only one neighbor, we know everything we need to know

    # Stage 2: Molecular Complex Prediction
    unvisited = set(edges)
    num_clusters = 0
    for seed in sorted(weights, key=weights.get, reverse=True):
        if seed in unvisited:
            cluster, frontier = {seed}, {seed}
            w = weights[seed] * weight_threshold
            while frontier:
                cluster.update(frontier)
                unvisited -= frontier
                frontier = set(
                    n
                    for n in set.union(*(edges[n] for n in frontier)) & unvisited
                    if weights[n] > w
                )

            # haircut: only keep 2-core complexes
            invalid_nodes = True
            while invalid_nodes and cluster:
                invalid_nodes = set(n for n in cluster if len(edges[n] & cluster) < 2)
                cluster -= invalid_nodes

            if cluster:
                res.append(list(cluster))
                num_clusters += 1
    return res


def __mcode_weighted(g, wlabel, weight_threshold):
    res = []
    graph = defaultdict(set)  # node id => neighboring node ids
    edges = defaultdict(lambda: defaultdict(int))

    # read in graph
    for a, b, data in g.edges(data=True):
        graph[a].add(b)
        graph[b].add(a)
        w = data[wlabel]
        edges[a][b] = w
        edges[b][a] = w

    # Stage 1: Vertex Weighting
    weights = dict((v, sum(edges[v].values()) / len(edges[v]) ** 2) for v in graph)
    for i, v in enumerate(graph):

        neighborhood = {v} | graph[v]
        # valid k-core with the highest weight
        if len(neighborhood) > 2:
            k = 2  # already covered k = 1
            while True:
                invalid_nodes = True
                while invalid_nodes and neighborhood:
                    invalid_nodes = set(
                        n for n in neighborhood if len(graph[n] & neighborhood) < k
                    )
                    neighborhood -= invalid_nodes
                if not neighborhood:
                    break

                # vertex weight = k-core number * density of k-core
                weights[v] = max(
                    weights[v],
                    k * sum(edges[v][n] for n in neighborhood) / len(neighborhood) ** 2,
                )
                k += 1

        # if node has only one neighbor, we know everything we need to know

    # Stage 2: Molecular Complex Prediction
    unvisited = set(graph)
    num_clusters = 0
    for seed in sorted(weights, key=weights.get, reverse=True):
        if seed in unvisited:
            cluster, frontier = {seed}, {seed}
            w = weights[seed] * weight_threshold
            while frontier:
                cluster.update(frontier)
                unvisited -= frontier
                frontier = set(
                    n
                    for n in set.union(*(graph[n] for n in frontier)) & unvisited
                    if weights[n] > w
                )

            # haircut: only keep 2-core complexes
            invalid_nodes = True
            while invalid_nodes and cluster:
                invalid_nodes = set(n for n in cluster if len(graph[n] & cluster) < 2)
                cluster -= invalid_nodes

            if cluster:
                res.append(list(cluster))
                num_clusters += 1
                if not unvisited:
                    break  # quit if all nodes visited
    return res



# -*- coding: utf-8 -*-
#
#    Copyright (C) 2018 by
#    Thomas Bonald <thomas.bonald@telecom-paristech.fr>
#    Bertrand Charpentier <bertrand.charpentier@live.fr>
#    All rights reserved.
#    BSD license.

import numpy as np
import networkx as nx


def paris(G, copy_graph=True):
    n = G.number_of_nodes()
    if copy_graph:
        F = G.copy()
    else:
        F = G

    # index nodes from 0 to n - 1
    if set(F.nodes()) != set(range(n)):
        F = nx.convert_node_labels_to_integers(F)

    # node weights
    w = {u: 0 for u in range(n)}
    wtot = 0
    for (u, v) in F.edges():
        if "weight" not in F[u][v]:
            F[u][v]["weight"] = 1
        weight = F[u][v]["weight"]
        w[u] += weight
        w[v] += weight
        wtot += weight
        if u != v:
            wtot += weight

    # cluster sizes
    s = {u: 1 for u in range(n)}

    # connected components
    cc = []

    # dendrogram as list of merges
    D = []

    # cluster index
    u = n
    while n > 0:
        # nearest-neighbor chain
        chain = [list(F.nodes())[0]]
        while len(chain) != 0:
            a = chain.pop()
            # nearest neighbor
            dmin = float("inf")
            b = -1
            for v in F.neighbors(a):
                if v != a:
                    d = w[v] * w[a] / float(F[a][v]["weight"]) / float(wtot)
                    if d < dmin:
                        b = v
                        dmin = d
                    elif d == dmin:
                        b = min(b, v)
            d = dmin
            if len(chain) != 0:
                c = chain.pop()
                if b == c:
                    # merge a,b
                    D.append([a, b, d, s[a] + s[b]])
                    # update graph
                    F.add_node(u)
                    neighbors_a = list(F.neighbors(a))
                    neighbors_b = list(F.neighbors(b))
                    for v in neighbors_a:
                        F.add_edge(u, v, weight=F[a][v]["weight"])
                    for v in neighbors_b:
                        if F.has_edge(u, v):
                            F[u][v]["weight"] += F[b][v]["weight"]
                        else:
                            F.add_edge(u, v, weight=F[b][v]["weight"])
                    F.remove_node(a)
                    F.remove_node(b)
                    n -= 1
                    # update weight and size
                    w[u] = w.pop(a) + w.pop(b)
                    s[u] = s.pop(a) + s.pop(b)
                    # change cluster index
                    u += 1
                else:
                    chain.append(c)
                    chain.append(a)
                    chain.append(b)
            elif b >= 0:
                chain.append(a)
                chain.append(b)
            else:
                # remove the connected component
                cc.append((a, s[a]))
                F.remove_node(a)
                w.pop(a)
                s.pop(a)
                n -= 1

    # add connected components to the dendrogram
    a, s = cc.pop()
    for b, t in cc:
        s += t
        D.append([a, b, float("inf"), s])
        a = u
        u += 1

    return __reorder_dendrogram(np.array(D))


def __reorder_dendrogram(D):
    n = np.shape(D)[0] + 1
    order = np.zeros((2, n - 1), float)
    order[0] = range(n - 1)
    order[1] = np.array(D)[:, 2]
    index = np.lexsort(order)
    nindex = {i: i for i in range(n)}
    nindex.update({n + index[t]: n + t for t in range(n - 1)})
    return np.array(
        [
            [nindex[int(D[t][0])], nindex[int(D[t][1])], D[t][2], D[t][3]]
            for t in range(n - 1)
        ]
    )[index, :]


# Rank clusterings at every level of the dendrogram
def __rank_clustering(D):
    logdist = np.log(D[:, 2])
    delta = logdist[1:] - logdist[:-1]
    return np.argsort(-delta[len(delta) // 2 :]) + 1 + len(delta) // 2


# Select the k-th best clustering
def paris_best_clustering(D, k=0):
    return __select_clustering(D, __rank_clustering(D)[k])


# Select the clustering after k merges
def __select_clustering(D, k):
    n = np.shape(D)[0] + 1
    k = min(k, n - 1)
    cluster = {i: [i] for i in range(n)}
    for t in range(k):
        cluster[n + t] = cluster.pop(int(D[t][0])) + cluster.pop(int(D[t][1]))
    return sorted(cluster.values(), key=len, reverse=True)



from cdlib.classes.clustering import Clustering
import networkx as nx

try:
    import igraph as ig
except ModuleNotFoundError:
    ig = None

from collections import defaultdict


class EdgeClustering(Clustering):
    """Edge Clustering representation.

    :param communities: list of communities
    :param graph: a networkx/igraph object
    :param method_name: community discovery algorithm name
    :param method_parameters: configuration for the community discovery algorithm used
    :param overlap: boolean, whether the partition is overlapping or not
    """

    def __init__(
        self,
        communities: list,
        graph: object,
        method_name: str = "",
        method_parameters: dict = None,
        overlap: bool = False,
    ):
        super().__init__(communities, graph, method_name, method_parameters, overlap)

        if graph is not None:
            edge_count = len(
                {eid: None for community in communities for eid in community}
            )
            if isinstance(graph, nx.Graph):
                self.node_coverage = edge_count / graph.number_of_edges()
            elif ig is not None and isinstance(graph, ig.Graph):
                self.node_coverage = edge_count / graph.ecount()
            else:
                raise ValueError("Unsupported Graph type.")

    def to_edge_community_map(self) -> dict:
        """
        Generate a <edge, list(communities)> representation of the current clustering

        :return: dict of the form <edge, list(communities)>
        """

        edge_to_communities = defaultdict(list)
        for cid, community in enumerate(self.communities):
            for edge in community:
                edge_to_communities[edge].append(cid)

        return edge_to_communities



import networkx as nx

import time
from random import random


class ModularityRCommunityDiscovery(object):
    minimum_improvement = 0.000001

    def __init__(self, graph):
        self.graph = graph
        self.starting_node = None
        self.community = []
        self.boundary = set()
        self.shell = set()
        self.remove_self_loops()

    def reset(self):
        self.community.clear()
        self.boundary.clear()
        self.shell.clear()

    def remove_self_loops(self):
        for node in self.graph.nodes():
            if self.graph.has_edge(node, node):
                self.graph.remove_edge(node, node)

    def set_start_node(self, start_node):
        if start_node in self.graph.nodes():
            self.starting_node = start_node
            self.community.append(start_node)
            self.boundary.add(start_node)
            self.shell = set(self.graph.neighbors(start_node))
        else:
            print("Invalid starting node! Try with another one.")
            exit(-1)

    def update_sets_when_node_joins(self, node, change_boundary=False):
        self.community.append(node)
        if change_boundary:
            self.update_boundary_when_node_joins(node)
        self.update_shell_when_node_joins(node)

    def update_shell_when_node_joins(self, new_node):
        self.shell.update(self.graph.neighbors(new_node))
        for node in self.community:
            self.shell.discard(node)

    def update_boundary_when_node_joins(self, new_node):
        should_be_boundary = False
        for neighbor in self.graph.neighbors(new_node):
            if (neighbor in self.community) is False:
                should_be_boundary = True
                break
        if should_be_boundary:
            self.boundary.add(new_node)

    def find_best_next_node(self, improvements):
        best_candidate = None
        best_improvement = -float("inf")
        for candidate, improvement in sorted(
            improvements.items(), key=lambda x: random()
        ):
            if improvement > best_improvement:
                best_candidate = candidate
                best_improvement = improvement
        return best_candidate

    def community_search(self, start_node):
        self.set_start_node(start_node)
        modularity_r = 0.0
        T = self.graph.degree[start_node]

        while (
            len(self.community) < self.graph.number_of_nodes() and len(self.shell) > 0
        ):
            delta_r = (
                {}
            )  # key: candidate nodes from the shell set, value: total improved strength after a node joins.
            delta_T = (
                {}
            )  # key: candidate nodes from the shell set, value: delta T (based on notations of the paper).
            for node in self.shell:
                delta_r[node], delta_T[node] = self.compute_modularity(
                    (modularity_r, T), node
                )

            new_node = self.find_best_next_node(delta_r)
            if delta_r[new_node] < ModularityRCommunityDiscovery.minimum_improvement:
                break

            modularity_r += delta_r[new_node]
            T += delta_T[new_node]
            self.update_sets_when_node_joins(new_node, change_boundary=True)

        return sorted(
            self.community
        )  # sort is only for a better representation, can be ignored to boost performance.

    def compute_modularity(self, auxiliary_info, candidate_node):
        R, T = auxiliary_info
        x, y, z = 0, 0, 0
        for neighbor in self.graph.neighbors(candidate_node):
            if neighbor in self.boundary:
                x += 1
            else:
                y += 1

        for neighbor in [
            node
            for node in self.graph.neighbors(candidate_node)
            if node in self.boundary
        ]:
            if self.should_leave_boundary(neighbor, candidate_node):
                for node in self.graph.neighbors(neighbor):
                    if (node in self.community) and ((node in self.boundary) is False):
                        z += 1
        return float(x - R * y - z * (1 - R)) / float(T - z + y), -z + y

    def should_leave_boundary(self, possibly_leaving_node, neighbor_node):
        neighbors = set(self.graph.neighbors(possibly_leaving_node))
        neighbors.discard(neighbor_node)
        for neighbor in neighbors:
            if (neighbor in self.community) is False:
                return False
        return True



from cdlib.classes.node_clustering import NodeClustering
import json


class FuzzyNodeClustering(NodeClustering):
    """Fuzzy Node Communities representation.

    :param communities: list of communities
    :param node_allocation: dictionary specifying for each node the allocation of probability toward the communities it is placed in
    :param graph: a networkx/igraph object
    :param method_name: community discovery algorithm name
    :param method_parameters: configuration for the community discovery algorithm used
    :param overlap: boolean, whether the partition is overlapping or not
    """

    def __init__(
        self,
        communities: list,
        node_allocation: dict,
        graph: object,
        method_name: str = "",
        method_parameters: dict = None,
        overlap: bool = False,
    ):
        super().__init__(communities, graph, method_name, method_parameters, overlap)
        self.allocation_matrix = node_allocation

    def to_json(self) -> str:
        """
        Generate a JSON representation of the algorithms object

        :return: a JSON formatted string representing the object
        """

        partition = {
            "communities": self.communities,
            "algorithm": self.method_name,
            "params": self.method_parameters,
            "overlap": self.overlap,
            "coverage": self.node_coverage,
            "allocation_matrix": self.allocation_matrix,
        }

        return json.dumps(partition)



from collections import namedtuple

import networkx as nx
from typing import Callable
import cdlib
from cdlib.evaluation.internal.TOPSIS import topsis
from cdlib.evaluation.internal.statistical_ranking import (
    bonferroni_dunn_test,
    friedman_test,
)
import numpy as np

__all__ = ["elem", "FitnessRanking"]

elem = namedtuple("elem", "rk alg param score")
elem.__new__.__defaults__ = (None,) * len(elem._fields)

post_hoc = namedtuple("post_hoc", "comparison z_value p_value adj_p_value")


class FitnessRanking(object):
    def __init__(self, graph: nx.Graph, partitions: list):
        """
        Class that provides functionalities for the generation of Clustering objects w.r.t. a given set of fitness scores.

        :param graph: a Networkx/iGraph object
        :param partitions: a list of Clustering objects
        """
        self.partitions = partitions
        self.graph = graph
        self.rankings = {}
        self.ranks = []
        self.rnk = {}

    def rank(
        self, scoring_function: Callable[[nx.Graph, object], object]
    ) -> [str, dict]:
        """
        Computes the specified scoring function to all the Clustering objects for which a ranking is required.

        :param scoring_function: a fitness function from cdlib.evaluation
        :return: a tuple whose first element is the scoring_function name, while the second is a dictionary having as key the clustering name and as value the computed score.

        :Example:

        >>> import networkx as nx
        >>> from cdlib import evaluation
        >>> from cdlib import algorithms
        >>> g = nx.karate_club_graph()
        >>> coms = algorithms.louvain(g)
        >>> coms2 = algorithms.demon(g, 0.25)
        >>> coms3 = algorithms.label_propagation(g)
        >>> coms4 = algorithms.angel(g, 0.6)
        >>> rk = evaluation.FitnessRanking(g, [coms2, coms, coms3, coms4])
        >>> rk.rank(evaluation.fraction_over_median_degree)
        """

        ranks = {}
        for partition in self.partitions:
            s = scoring_function(self.graph, partition, summary=True)
            ranks[f"{partition.method_name}_{partition.method_parameters}"] = s.score
        s_ranks = sorted(ranks.items(), key=lambda x: -x[1])
        s_ranks = [
            elem(rk=x + 1, alg=c[0].split("_")[0], param=c[0].split("_")[1], score=c[1])
            for x, c in enumerate(s_ranks)
        ]

        res = {f"{i.alg}_{i.param}": (i.rk, i.score) for i in s_ranks}

        self.rankings[scoring_function.__name__] = res

        return scoring_function.__name__, ranks

    def topsis(self) -> [list, None]:
        """
        The Technique for Order of Preference by Similarity to Ideal Solution (TOPSIS) is a multi-criteria decision analysis method.
        TOPSIS is based on the concept that the chosen alternative should have the shortest geometric distance from the positive ideal solution (PIS) and the longest geometric distance from the negative ideal solution (NIS).

        :return: a tuple whose first element is the ranking dictionary assigning a TOPSIS score to each Clustering object, while the second is None (to maintain coherence with friedman_ranking).

        :Example:

        >>> import networkx as nx
        >>> from cdlib import evaluation
        >>> from cdlib import algorithms
        >>> g = nx.karate_club_graph()
        >>> coms = algorithms.louvain(g)
        >>> coms2 = algorithms.demon(g, 0.25)
        >>> coms3 = algorithms.label_propagation(g)
        >>> coms4 = algorithms.angel(g, 0.6)
        >>> rk = evaluation.FitnessRanking(g, [coms2, coms, coms3, coms4])
        >>> rk.rank(evaluation.fraction_over_median_degree)
        >>> rk.rank(evaluation.edges_inside)
        >>> rk.rank(evaluation.cut_ratio)
        >>> rk.rank(evaluation.erdos_renyi_modularity)
        >>> rk.rank(evaluation.newman_girvan_modularity)
        >>> rk.rank(evaluation.modularity_density)
        >>> rnk, _ = rk.topsis()

        :References:

        1. Hwang, C.L.; Yoon, K. (1981). Multiple Attribute Decision Making: Methods and Applications. New York: Springer-Verlag.
        2. Yoon, K. (1987). "A reconciliation among discrete compromise situations". Journal of the Operational Research Society. 38 (3): 277–286. doi:10.1057/jors.1987.44.
        """
        rp = {
            a: [] for score, value in self.rankings.items() for a, vals in value.items()
        }
        for score, value in self.rankings.items():
            for a, vals in value.items():
                rp[a].append(vals[1])
        ranks = np.array([np.array(x) for x in rp.values()])

        agg_rank = topsis(ranks, [1] * len(self.rankings), "v", "m")
        for ids, alg in enumerate(rp):
            rp[alg] = agg_rank[ids]

        s_ranks = sorted(rp.items(), key=lambda x: -x[1])
        self.ranks = [
            elem(rk=x + 1, alg=c[0].split("_")[0], param=c[0].split("_")[1], score=c[1])
            for x, c in enumerate(s_ranks)
        ]

        self.rnk = {f"{i.alg}_{i.param}": i.score for i in self.ranks}

        return self.ranks, None

    def friedman_ranking(self) -> [list, float]:
        """
        Performs a Friedman ranking test.
        Tests the hypothesis that in a set of k dependent samples groups (where k >= 2) at least two of the groups represent populations with different median values.


        :return: a tuple whose first element is a dictionary assigning a rank to each Clustering object, while the second is the p-value associated to the ranking.

        :Example:

        >>> import networkx as nx
        >>> from cdlib import evaluation
        >>> from cdlib import algorithms
        >>> g = nx.karate_club_graph()
        >>> coms = algorithms.louvain(g)
        >>> coms2 = algorithms.demon(g, 0.25)
        >>> coms3 = algorithms.label_propagation(g)
        >>> coms4 = algorithms.angel(g, 0.6)
        >>> rk = evaluation.FitnessRanking(g, [coms2, coms, coms3, coms4])
        >>> rk.rank(evaluation.fraction_over_median_degree)
        >>> rk.rank(evaluation.edges_inside)
        >>> rk.rank(evaluation.cut_ratio)
        >>> rk.rank(evaluation.erdos_renyi_modularity)
        >>> rk.rank(evaluation.newman_girvan_modularity)
        >>> rk.rank(evaluation.modularity_density)
        >>> rnk, p_value = rk.friedman_ranking()

        :References:

        1. M. Friedman, The use of ranks to avoid the assumption of normality implicit in the analysis of variance, Journal of the American Statistical Association 32 (1937) 674–701.
        2. D.J. Sheskin, Handbook of parametric and nonparametric statistical procedures. crc Press, 2003, Test 25: The Friedman Two-Way Analysis of Variance by Ranks
        """

        rp = {
            a: [] for score, value in self.rankings.items() for a, vals in value.items()
        }
        for score, value in self.rankings.items():
            for a, vals in value.items():
                rp[a].append(vals[1])
        ranks = [x for x in rp.values()]

        for ids, alg in enumerate(rp):
            rp[alg] = ranks[ids]

        statistic, p_value, ranking, rank_cmp = friedman_test(*rp.values())

        self.rnk = {key: rank_cmp[i] for i, key in enumerate(rp.keys())}

        s_ranks = sorted(self.rnk.items(), key=lambda x: -x[1])
        self.ranks = [
            elem(rk=x + 1, alg=c[0].split("_")[0], param=c[0].split("_")[1], score=c[1])
            for x, c in enumerate(s_ranks)
        ]

        return self.ranks, p_value

    def bonferroni_post_hoc(self) -> list:
        """
        Performs a Bonferroni-Dunn post-hoc test using the pivot quantities obtained by a ranking test.
        Tests the hypothesis that the ranking of the control method (best ranked clustering) is different to each of the other methods.

        :return: a list of named tuples reporting the pairwise statistical significant comparisons among the best ranked clustering and the others (in terms of z-value, p-value, adjusted-p-value)

        :Example:

        >>> import networkx as nx
        >>> from cdlib import evaluation
        >>> from cdlib import algorithms
        >>> g = nx.karate_club_graph()
        >>> coms = algorithms.louvain(g)
        >>> coms2 = algorithms.demon(g, 0.25)
        >>> coms3 = algorithms.label_propagation(g)
        >>> coms4 = algorithms.angel(g, 0.6)
        >>> rk = evaluation.FitnessRanking(g, [coms2, coms, coms3, coms4])
        >>> rk.rank(evaluation.fraction_over_median_degree)
        >>> rk.rank(evaluation.edges_inside)
        >>> rk.rank(evaluation.cut_ratio)
        >>> rk.rank(evaluation.erdos_renyi_modularity)
        >>> rk.rank(evaluation.newman_girvan_modularity)
        >>> rk.rank(evaluation.modularity_density)
        >>> rnk, p_value = rk.friedman_ranking()
        >>> pc = rk.bonferroni_post_hoc()

        :References:

        O.J. Dunn, Multiple comparisons among means, Journal of the American Statistical Association 56 (1961) 52–64.
        """

        res = []
        comparisons, z_values, p_values, adj_p_values = bonferroni_dunn_test(self.rnk)
        for i in range(0, len(comparisons)):
            res.append(
                post_hoc(
                    comparison=comparisons[i],
                    z_value=z_values[i],
                    p_value=p_values[i],
                    adj_p_value=adj_p_values[i],
                )
            )
        return res



from __future__ import division
import numpy as np
import os.path

"""
This module implements community detection.
"""

__author__ = ["Thomas Aynaud (thomas.aynaud@lip6.fr)"]
#    Copyright (C) 2009 by
#    Thomas Aynaud <thomas.aynaud@lip6.fr>
#    All rights reserved.
#    BSD license.

import networkx as nx


class ML2(object):
    __MIN = 0.000001
    __PASS_MAX = -1
    LOGOPERATIONS = False

    nbVertices = 0

    def __init__(self, graph, attributes, authorIndex):
        self.graph = graph
        self.graphBase = graph.copy()
        self.attributes = attributes
        self.nbVertices = len(graph)
        self.statusTab = []
        self.authorIndex = authorIndex

        # Build status structures
        status = Status()
        status.init(graph)
        self.statusTab.append(status)
        statusA = Status()
        statusA.initAttribStatus(graph, authorIndex, attributes)
        self.statusTab.append(statusA)

        self.status_list = list()

    def critereCombinaison(self):
        return (
            self.__modularity(self.statusTab[0]) + self.__modularity(self.statusTab[1])
        ) / 2.0

    def findPartition(self):
        giniMatrix = self.calculateGiniMatrixInitial()
        self.__one_level(giniMatrix=None)
        new_mod = self.critereCombinaison()

        partition, bijection = self.__renumber()

        self.status_list.append(partition)
        mod = new_mod
        self.induced_graph(partition)
        self.statusTab[0].init(self.graph)

        # if(args.verbose):
        #    print("Inducing attrib status")
        self.statusTab[1].inducedAttribStatusTab(partition, bijection)
        giniMatrix = self.firstInducedGiniMatrix(partition, giniMatrix)

        while True:
            self.__one_level(giniMatrix=giniMatrix)
            new_mod = self.critereCombinaison()
            if new_mod - mod < self.__MIN:
                break
            partition, bijection = self.__renumber()

            self.status_list.append(partition)

            mod = new_mod
            self.induced_graph(partition)
            giniMatrix = self.inducedGiniMatrix(partition, giniMatrix)

            self.statusTab[0].init(self.graph)

            if self.statusTab[1] != None:
                self.statusTab[1].inducedAttribStatusTab(partition, bijection)

        dendogram = self.status_list[:]

        # Generate and output partition
        partition = dendogram[0].copy()
        for index in range(1, len(dendogram)):
            for node, community in partition.items():
                partition[node] = dendogram[index][community]

        return partition

    def dist(self, v1, v2):
        attrV1 = self.attributes[v1]
        attrV2 = self.attributes[v2]
        distance = 0.0
        for attr, val1 in attrV1.items():
            val2 = attrV2.get(attr, 0.0)
            distance += (val1 - val2) ** 2
        for attr, val2 in attrV2.items():
            if not attr in attrV1:
                distance += val2 * val2
        return distance

    def distArray(self, v1, v2):
        attrV1 = self.attributes[v1]
        attrV2 = self.attributes[v2]
        distance = 0.0
        for i in range(len(attrV1)):
            distance += (attrV1[i] - attrV2[i]) ** 2
        return distance

    def firstInducedGiniMatrix(self, partition, giniMatrix):
        out = np.zeros([len(set(partition.values())), len(set(partition.values()))])
        # if(args.verbose):
        #    pprint(giniMatrix)
        for i in partition:
            for j in partition:
                out[partition[i]][partition[j]] = giniMatrix[self.authorIndex[i]][
                    self.authorIndex[j]
                ]
        return out

    def inducedGiniMatrix(self, partition, giniMatrix):
        # if(args.verbose):
        #    print("inducedGiniMatrix...")
        out = np.zeros([len(set(partition.values())), len(set(partition.values()))])
        for i in partition:
            for j in partition:
                out[partition[i]][partition[j]] = (
                    out[partition[i]][partition[j]] + giniMatrix[i][j]
                )
        # if(args.verbose):
        #    print("End inducedGiniMatrix")

        return out

    def calculateGiniMatrixInitial(self):
        giniMatrix = {}
        for v1 in self.graph:
            giniMatrix[self.authorIndex[v1]] = {}
        np.zeros(self.nbVertices ** 2).reshape((self.nbVertices, self.nbVertices))
        for v1 in self.graph:
            for v2 in self.graph:
                d = (
                    -1
                    * self.dist(self.authorIndex[v1], self.authorIndex[v2])
                    / self.nbVertices ** 2
                )
                giniMatrix[self.authorIndex[v1]][self.authorIndex[v2]] = d
                giniMatrix[self.authorIndex[v2]][self.authorIndex[v1]] = d
        """
        print "Calculating Gini Matrix Initial"
        Y = pdist(self.attributes, 'sqeuclidean')
        print "division"
        Y = np.divide(Y, 0.0-float(len(self.attributes)**2))
        pprint(giniMatrix)
        pprint(squareform(Y))
        return squareform(Y)
        """
        return giniMatrix

    def induced_graph(self, partition):
        newGraph = nx.Graph()
        newGraph.add_nodes_from(partition.values())

        # for node1, node2, datas in self.graph.edges_iter(data = True) :
        for node1, node2, datas in list(self.graph.edges(data=True)):
            weight = datas.get("weight", 1)
            com1 = partition[node1]
            com2 = partition[node2]
            w_prec = newGraph.get_edge_data(com1, com2, {"weight": 0}).get("weight", 1)
            newGraph.add_edge(com1, com2, weight=w_prec + weight)
        self.graph = newGraph

    def __renumber(self):
        count = 0
        dictionary = self.statusTab[0].node2com
        ret = dictionary.copy()
        new_values = dict([])
        for key in dictionary.keys():
            value = dictionary[key]
            new_value = new_values.get(value, -1)
            if new_value == -1:
                new_values[value] = count
                new_value = count
                count = count + 1
            ret[key] = new_value
        return ret, new_values

    def __one_level(self, giniMatrix=None):
        modif = True

        while modif:
            modif = False
            numNode = 0
            for node in self.graph.nodes():
                numNode = numNode + 1

                com_node = self.statusTab[0].node2com[node]
                best_com = com_node

                best_increase = 0
                neigh_communities = self.__neighcom(node, giniMatrix=giniMatrix)
                # if(args.verbose):
                #    print("Neighb Communities of " + str(node))
                #    pprint(neigh_communities)

                degc_totw_tab = []

                for i in range(len(self.statusTab)):

                    degc_totw_tab.append(
                        self.statusTab[i].gdegrees.get(node, 0.0)
                        / (self.statusTab[i].total_weight * 2.0)
                    )
                    theWeight = neigh_communities[com_node][i]

                    if abs(self.statusTab[i].degrees[com_node]) <= abs(
                        self.statusTab[i].gdegrees[node]
                    ):
                        self.statusTab[i].degrees[com_node] = abs(
                            self.statusTab[i].gdegrees[node]
                        )

                    self.__remove(node, com_node, theWeight, self.statusTab[i])
                assert (
                    self.statusTab[0].node2com[node] == self.statusTab[1].node2com[node]
                )

                # Find the best community
                for com, dnc in neigh_communities.items():
                    incr = 0.0
                    for i in range(len(self.statusTab)):
                        totw = abs(self.statusTab[i].total_weight)
                        if i == 0:
                            a = (
                                abs(dnc[i])
                                - abs(
                                    self.statusTab[i].degrees.get(com, 0.0)
                                    * degc_totw_tab[i]
                                )
                            ) / totw
                            incr += a
                        else:
                            a = (
                                0.0
                                - abs(dnc[i])
                                + abs(
                                    self.statusTab[i].degrees.get(com, 0.0)
                                    * degc_totw_tab[i]
                                )
                            ) / totw
                            incr += a
                    incr /= 2
                    if incr > best_increase:
                        best_increase = incr
                        best_com = com

                for i in range(len(self.statusTab)):
                    if best_com in neigh_communities:
                        theWeight = neigh_communities[best_com][i]
                    else:
                        print(
                            "IS THAT POSSIBLE ???? (best_com not in neigh_communities)"
                        )
                        exit(0)
                        theWeight = 0
                    self.__insert(node, best_com, theWeight, self.statusTab[i])

                if best_com != com_node:
                    modif = True

    def __neighcom(self, node, giniMatrix=None):
        weights = {}
        voisins = self.graph[node].items()
        curCommunity = self.statusTab[0].node2com[node]
        if curCommunity not in weights:
            weights[curCommunity] = np.zeros([len(self.statusTab)])

        for neighbor, datas in voisins:
            if neighbor != node:
                weight = datas.get("weight", 1)
                neighborcom = self.statusTab[0].node2com[neighbor]

                if neighborcom not in weights:
                    weights[neighborcom] = np.zeros([len(self.statusTab)])

                # For the graph
                weights[neighborcom][0] = weights[neighborcom][0] + weight

                # For the attributes
                if giniMatrix is not None:
                    weight = giniMatrix[node][neighbor]
                else:
                    weight = (
                        -1
                        * self.dist(self.authorIndex[node], self.authorIndex[neighbor])
                        / self.nbVertices ** 2
                    )
                    weights[neighborcom][1] = weights[neighborcom][1] + weight
        return weights

    def __remove(self, node, com, weight, status):
        status.degrees[com] = status.degrees.get(com, 0.0) - status.gdegrees.get(
            node, 0.0
        )
        status.internals[com] = float(
            status.internals.get(com, 0.0) - weight - status.loops.get(node, 0.0)
        )
        status.node2com[node] = -1

    def __insert(self, node, com, weight, status):
        status.node2com[node] = com
        status.degrees[com] = status.degrees.get(com, 0.0) + status.gdegrees.get(
            node, 0.0
        )
        status.internals[com] = float(
            status.internals.get(com, 0.0) + weight + status.loops.get(node, 0.0)
        )

    def __modularity(self, status):
        links = abs(float(status.total_weight))
        result = 0.0
        for community in set(status.node2com.values()):
            in_degree = abs(status.internals.get(community, 0.0))
            degree = abs(status.degrees.get(community, 0.0))
            expected = (degree / (2.0 * links)) ** 2
            found = in_degree / links
            if status.total_weight < 0:
                result += expected - found
            else:
                result += found - expected
        return result


class Status(object):
    """
    To handle several data in one struct.
    Could be replaced by named tuple, but don't want to depend on python 2.6
    """

    node2com = dict([])
    total_weight = 0
    internals = dict([])
    degrees = dict([])
    gdegrees = dict([])
    loops = dict([])

    def __str__(self):
        return (
            "------------------------\nnode2com : "
            + str(self.node2com)
            + "\n degrees : "
            + str(self.degrees)
            + "\n gdegrees : "
            + str(self.gdegrees)
            + "\n internals : "
            + str(self.internals)
            + "\n total_weight : "
            + str(self.total_weight)
            + "\n loops:"
            + str(self.loops)
            + "\n-----------------------"
        )

    def initAttribStatus(self, graph, authorIndex, attributes):
        """Initialize the status of an attributes list with every node in one community"""
        N = len(graph)
        count = 0

        # Compute the center of gravity using dict
        meanVector = {}
        for v, attrs in attributes.items():
            for attrId, attrValue in attrs.items():
                meanVector[attrId] = meanVector.get(attrId, 0.0) + attrValue
        for attrId, attrValue in meanVector.items():
            meanVector[attrId] = meanVector[attrId] / N

        variance = {}
        for node in sorted(graph.nodes()):
            for attrId, attrValue in meanVector.items():
                variance[attrId] = variance.get(attrId, 0.0) + (
                    (attrValue - attributes[authorIndex[node]].get(attrId, 0.0)) ** 2
                )
        inertieTot = 0.0
        for v in variance.values():
            inertieTot += v / N

        self.total_weight = 0.0 - inertieTot

        for node in sorted(graph.nodes()):
            self.node2com[node] = count

            # Compute the distance to the center of gravity
            distanceToCenterOfGravity = 0.0
            for attrId, attrValue in meanVector.items():
                distanceToCenterOfGravity += (
                    attrValue - attributes[authorIndex[node]].get(attrId, 0.0)
                ) ** 2

            phiHuyghens = -1 * (inertieTot + distanceToCenterOfGravity) / N
            # if(args.verbose):
            #    print("# phiHuyghens(" + str(node) + ") = " + str(phiHuyghens))
            self.degrees[count] = phiHuyghens
            self.gdegrees[node] = phiHuyghens
            self.loops[node] = 0
            self.internals[count] = self.loops[node]
            count = count + 1

    def inducedAttribStatusTab(self, node2com, bijection):
        # if(args.verbose):
        #    print(self)
        retrobijection = {}
        for k, v in bijection.items():
            retrobijection[v] = k
        self.node2com = dict([])
        oldDegrees = self.degrees
        oldInternals = self.internals

        self.degrees = dict([])
        self.gdegrees = dict([])
        self.internals = dict([])
        self.node2com = dict([])
        self.loops = dict([])

        for node in retrobijection:
            self.node2com[node] = node
            deg = oldDegrees[retrobijection[node]]
            self.degrees[node] = deg
            self.gdegrees[node] = deg
            self.loops[node] = oldInternals[retrobijection[node]]
            self.internals[node] = self.loops[node]

    def init(self, graph):
        """Initialize the status of a graph with every node in one community"""
        count = 0
        self.node2com = dict([])
        self.degrees = dict([])
        self.gdegrees = dict([])
        self.internals = dict([])
        self.total_weight = graph.size(weight="weight")
        for node in sorted(graph.nodes()):
            self.node2com[node] = count
            deg = float(graph.degree(node, weight="weight"))
            self.degrees[count] = deg
            self.gdegrees[node] = deg
            self.loops[node] = float(
                graph.get_edge_data(node, node, {"weight": 0}).get("weight", 1)
            )
            self.internals[count] = self.loops[node]
            count = count + 1


def loadDataset(path):
    graph = nx.Graph()

    # Read the graph
    if not os.path.isfile(path + ".edgeList"):
        print("Error: file '" + path + ".edgeList' not found")
        exit(-1)
    with open(path + ".edgeList") as f:
        for line in f.readlines():
            v1 = int(line.split(" ")[0])
            v2 = int(line.split(" ")[1])
            graph.add_node(v1)
            graph.add_node(v2)
            graph.add_edge(v1, v2)

    # Read the attributes
    attributes = {}
    for n in graph:
        attributes[n] = {}

    if not os.path.isfile(path + ".attributes"):
        print("Error: file '" + path + ".attributes' not found")
        exit(-1)

    with open(path + ".attributes") as f:
        for line in f.readlines():
            vertexId = int(line.split(" ")[0])
            elems = line.split(" ")[1].split(",")
            i = 0
            attrValues = {}
            for attrValue in elems:
                attrValues[i] = float(attrValue)
                i = i + 1
            attributes[vertexId] = attrValues

    # Build authorIndex
    authorIndex = {}
    for n in graph:
        authorIndex[n] = n

    # if(args.verbose):
    #    print("# Finished reading dataset")
    if os.path.exists(path + ".2ModLouvain"):
        os.remove(path + ".2ModLouvain")

    return graph, attributes, authorIndex



from .node_clustering import NodeClustering
from .edge_clustering import EdgeClustering
from .fuzzy_node_clustering import FuzzyNodeClustering
from .attr_node_clustering import AttrNodeClustering
from .bipartite_node_clustering import BiNodeClustering
from .temporal_clustering import TemporalClustering
from .named_clustering import NamedClustering



import numpy as np
import math
import functools

# https://github.com/RapidsAtHKUST/CommunityDetectionCodes


class FuncTag:
    def __init__(self):
        pass

    exp_inv_mul_tag = "exp_inv_mul"
    mul_tag = "mul"
    min_tag = "min"
    max_tag = "max"


def get_coefficient_func(tag: str) -> float:
    if tag == FuncTag.exp_inv_mul_tag:
        return lambda l, r: 1.0 / functools.reduce(
            lambda il, ir: il * ir, map(lambda ele: 1.0 + math.exp(2 - ele), [l, r]), 1
        )
    elif tag == FuncTag.mul_tag:
        return lambda l, r: l * r
    elif tag == FuncTag.min_tag:
        return min
    elif tag == FuncTag.max_tag:
        return max


def cal_modularity(input_graph, comm_result):
    return LinkBelongModularity(
        input_graph, comm_result, get_coefficient_func(FuncTag.exp_inv_mul_tag)
    ).calculate_modularity()


class LinkBelongModularity:
    PRECISION = 0.0001

    def __init__(self, input_graph, comm_result, coefficient_func):
        """
        :type input_graph: nx.Graph
        """
        self.comm_list = comm_result
        self.graph = input_graph
        self.coefficient_func = coefficient_func
        self.belong_weight_dict = {}
        self.in_degree_dict = {}
        self.out_degree_dict = {}

        self.init_belong_weight_dict()
        self.init_degree_dicts()

    def init_belong_weight_dict(self):
        belong_dict = {}
        for comm in self.comm_list:
            for mem in comm:
                if mem not in belong_dict:
                    belong_dict[mem] = 0
                belong_dict[mem] += 1
        for mem in belong_dict:
            self.belong_weight_dict[mem] = (
                1.0 / belong_dict[mem] if belong_dict[mem] != 0 else 0
            )

    def init_degree_dicts(self):
        for vertex in self.graph.nodes():
            # since graph here studied are used in undirected manner
            self.in_degree_dict[vertex] = self.graph.degree(vertex)
            self.out_degree_dict[vertex] = self.graph.degree(vertex)

    def calculate_modularity(self) -> float:
        modularity_val = 0
        vertex_num = self.graph.number_of_nodes()
        edge_num = self.graph.number_of_edges()
        for comm in self.comm_list:
            comm_size = len(comm)
            f_val_matrix = np.ndarray(shape=(comm_size, comm_size), dtype=float)
            f_val_matrix.fill(0)
            f_sum_in_vec = np.zeros(comm_size, dtype=float)
            f_sum_out_vec = np.zeros(comm_size, dtype=float)
            in_deg_vec = np.zeros(comm_size, dtype=float)
            out_deg_vec = np.zeros(comm_size, dtype=float)

            # calculate f_val_matrix, f_sum_in, f_sum_out
            for i in range(comm_size):
                src_mem = comm[i]
                in_deg_vec[i] = self.in_degree_dict[src_mem]
                out_deg_vec[i] = self.out_degree_dict[src_mem]
                for j in range(comm_size):
                    dst_mem = comm[j]
                    if i != j and self.graph.has_edge(src_mem, dst_mem):
                        f_val_matrix[i][j] = self.coefficient_func(
                            self.belong_weight_dict[src_mem],
                            self.belong_weight_dict[dst_mem],
                        )
                        f_sum_out_vec[i] += f_val_matrix[i][j]
                        f_sum_in_vec[j] += f_val_matrix[i][j]

            f_sum_in_vec /= vertex_num
            f_sum_out_vec /= vertex_num

            for i in range(comm_size):
                for j in range(comm_size):
                    if i != j and f_val_matrix[i][j] > LinkBelongModularity.PRECISION:
                        null_model_val = (
                            out_deg_vec[i]
                            * in_deg_vec[j]
                            * f_sum_out_vec[i]
                            * f_sum_in_vec[j]
                            / edge_num
                        )
                        modularity_val += f_val_matrix[i][j] - null_model_val
        modularity_val /= edge_num
        return modularity_val



# -*- coding: utf-8 -*-
"""
Created on Wed Aug 12 09:14:20 2020
@author: saipr
"""
import pandas as pd
import networkx as nx
import numpy as np
import matplotlib.pyplot as plt
from networkx import community as c
import community as com
import itertools
import random
from random import sample
from networkx import algorithms as al


# --------------------Methods in Core Expansion--------------------------------#

# Managing method to call everything as need be
def findCommunities(g, tol=0.0005):
    # assign all the edges their neighborhoodOverlap
    nx.set_edge_attributes(g, computeNeighborOverlap(g), "neighborOverlap")
    # assign all the nodes their sumOverlap attributes
    nx.set_node_attributes(g, computeSumOverlap(g), "sumOverlap")
    # assign every node a coreVal of -1
    setup = list()
    setup.append(-1)
    nx.set_node_attributes(g, setup, "coreValue")

    # find all the local maximums in sumOverlap
    communities = findCores(g)
    # Assign initial cores their unique core value
    for xx in range(0, len(communities)):
        for co in communities[xx]:
            g.nodes[co]["coreValue"] = [xx]

            # sweep 1
    iterr = 0
    while True:
        changed = False
        nodeIter = iter(g.nodes)
        coreValDict = nx.get_node_attributes(g, "coreValue")
        # now go through all the nodes and edit the unassigned
        for i in range(0, len(g.nodes)):
            check = next(nodeIter)
            if coreValDict[check][0] == -1:  # if the node is unassigned
                test = findClosestCore(communities, check, g)
                if test[0] != -1 and (len(test) < (tol * len(communities))):
                    coreValDict.update({check: test})
                    changed = True
        iterr += 1

        # if something has changed, update the list
        if changed:
            nx.set_node_attributes(g, coreValDict, "coreValue")  # update core values
            # the core list is updated, let's update the community list
            nodeIter = iter(g.nodes)  # for each node in the community
            for j in range(0, len(g.nodes)):
                thisOne = next(nodeIter)  # what node we're checking
                temppp = g.nodes[thisOne]["coreValue"]
                for thisCore in temppp:
                    # check if its coreval is not negative one
                    if thisCore != -1:
                        # if it's not negative, check if it's not already in that community
                        if thisOne in communities[thisCore]:
                            pass
                        else:  # and if not, add it in
                            communities[thisCore].append(thisOne)
        else:  # if nothing's changed, break the loop
            break

    # now do the second sweep
    nodeIter = iter(g.nodes)
    coreValDict = nx.get_node_attributes(g, "coreValue")

    for i in range(0, len(g.nodes)):
        straggler = next(nodeIter)
        coreValsTemp = g.nodes[straggler]["coreValue"]
        if coreValsTemp[0] == -1:  # if the node is unassigned
            byConnections = secondSort(
                g, straggler, communities
            )  # check if it belongs to any core by # of connections
            coreValDict.update({straggler: byConnections})
            for b in byConnections:
                if b != -1:
                    communities[b].append(straggler)

    nx.set_node_attributes(g, coreValDict, "coreValue")

    # step 6: print out any unassigned nodes, and what proportion they are out of all of them
    areUnassigned(g)

    return communities


# compute neighborhood overlaps- takes in a graph, returns a dictionary to set the values of g.edges with
def computeNeighborOverlap(g):
    # creating a dictionary of edges and thier neighborhood overlap value
    edgeOverlaps = {}
    # to iterate over the edges in g
    edgeIter = iter(g.edges)  # next() returns a touple with to and from
    for i in range(0, len(g.edges)):
        thisEdge = next(edgeIter)
        overVal = -1  # current overlap value
        to = thisEdge[0]  # first element of the edge touple, to
        fromm = thisEdge[1]  # second element of the edge touple, from

        # calculate the edgeOverlap between the two and from of thisEdge
        # cuv / (ku + kv - 2 - cuv)
        cuv = sum(
            1 for e in nx.common_neighbors(g, to, fromm)
        )  # shared neighbors of the endpoints
        ku = sum(1 for e in nx.all_neighbors(g, to))
        kv = sum(1 for e in nx.all_neighbors(g, fromm))
        overVal = 0
        try:
            overVal = cuv / (ku + kv - 2 - cuv)
        except:
            pass
        edgeOverlaps.update({thisEdge: overVal})
    return edgeOverlaps


# compute a node's sumOverlap - takes a graph, returns a directory to set the values of g.nodes with
def computeSumOverlap(g):
    nodeIter = iter(g.nodes)  # our darling node iteraor
    sumOverlaps = {}

    # for each node in G
    for i in range(0, len(g.nodes)):
        thisNode = next(nodeIter)  # the current node we're focusing on
        summ = 0  # curent sum neighbor overlap
        thisNodesEdges = g.edges(thisNode)  # we need the edges of this node
        thisEdgeIter = iter(thisNodesEdges)  # and to make an iterator out of them
        # add the sum of its edges' neighbor overlaps
        try:
            while True:
                thisEdge = next(thisEdgeIter)
                summ += g.edges[thisEdge]["neighborOverlap"]
        except:
            pass
        sumOverlaps.update({thisNode: summ})
    return sumOverlaps


# find the cores of a graph - nodes with a local maximum among their peers in sumOverlap
# also assigns the cores their core index, changing it away from -1 (so unassigned are nodes with core value -1)
def findCores(g):
    # given a graph g
    # create a list 'cores' to return
    cores = list()

    # for each node in graph g
    nodeIter = iter(g.nodes)
    for i in range(0, len(g.nodes)):
        passs = False  # should we skip this node
        thisNode = next(nodeIter)
        # get all their neighbors and how many there are
        numNeigh = sum(1 for i in nx.neighbors(g, thisNode))
        neighIter = nx.neighbors(g, thisNode)

        # check if any of their neighbors are already in the core list (if so, this node is NOT the local maximum)
        for j in range(0, numNeigh):
            m = next(neighIter)
            for c in cores:
                for q in range(0, len(c)):
                    if m == c[q]:
                        passs = True

                    # else, check if this node is the local maximum
        if not passs:
            # turn the sumOverlap of all their neighbors into a list
            check = g.nodes[thisNode]["sumOverlap"]  # what we're checking for
            neighIter = nx.neighbors(
                g, thisNode
            )  # reset the neighbor iterator to the top
            againstThese = {}
            for z in range(0, numNeigh):
                thisOne = next(neighIter)
                againstThese.update({thisOne: g.nodes[thisOne]["sumOverlap"]})

            # check if their sumOverlap is higher than all of their neighbors' (np.all())
            if check > np.all(np.array(list(againstThese.values()))):
                tem = list()
                tem.append(thisNode)
                cores.append(tem)
            elif check >= np.all(
                np.array(list(againstThese.values()))
            ):  # there's a speed tie
                ties = list()
                ties.append(thisNode)
                for key, value in againstThese.items():
                    if (value == check) and (key != thisNode):
                        ties.append(key)
                cores.append(ties)  # returns a group of cores that compose the same mx
    return cores


# given a graph and a check node, returns the core value of the highest neighbor of the graph
def findClosestCore(communities, n, g):
    coreID = list()
    currentMax = -999
    ties = 0
    for c in range(0, len(communities)):
        temp = 0
        for u, v, dictt in g.edges(n, data=True):  # iterate through all the edges in n
            if n == u:  # v is the other node
                if (
                    v in communities[c]
                ):  # if the other node is in the community in question
                    temp += dictt["neighborOverlap"]  # update comm's temp score
            else:  # u is the other node
                if u in c:  # if the other node is in the community in question
                    temp += dictt["neighborOverlap"]  # update comm's temp score
        if (temp > currentMax) and not (c in coreID):
            currentMax = temp
            coreID.clear()
            ties = 0
            coreID.append(c)
        elif temp == currentMax:
            ties += 1
            coreID.append(c)
    if len(coreID) == 0:
        coreID.append(-1)
    return coreID


# assign the stragglers by number of connections, not quality of connections
def secondSort(g, straggler, communities):
    tally = np.array([0] * len(communities))
    neighIter = iter(nx.neighbors(g, straggler))
    while True:
        try:
            this = next(neighIter)
        except:
            break
        temp = g.nodes[this]["coreValue"]
        for core in range(0, len(temp)):
            mark = temp[core]
            if mark != -1:
                tally[mark] += 1
    if np.any(tally) > 0:
        # find the max and assign them to their max communities
        maxx = list()  # records the indexes of the max values
        maxx.append(-1)
        for m in range(0, len(tally)):
            if tally[m] > tally[maxx[0]]:
                maxx.clear()
                maxx.append(m)
            elif tally[m] == tally[maxx[0]]:
                maxx.append(m)
        return maxx
    return [-1]


# checks if any node is unassigned, what node it is
def areUnassigned(g):
    nodeIter = iter(g.nodes)
    count = 0
    numOverlapping = 0
    for i in range(0, len(g.nodes)):
        this = next(nodeIter)
        tempp = g.nodes[this]["coreValue"]
        if tempp[0] == -1:
            count += 1
        else:
            if len(tempp) > 1:
                numOverlapping += 1
    return None


# Function to convert communities lists to node dictionary - for Communities API
def convertToCommunityDictionary(communities):
    communityDictionary = dict()
    index = 0
    for comm in communities:
        for co in comm:
            communityDictionary.update({co: index})
        index += 1
    return communityDictionary



import scipy as sp
import numpy as np


def friedman_test(*args: dict) -> [float, float, list, list]:
    """
    Performs a Friedman ranking test.
    Tests the hypothesis that in a set of k dependent samples groups (where k >= 2) at least two of the groups represent populations with different median values.

    Parameters
    ----------
    sample1, sample2, ... : array_like
        The sample measurements for each group.

    Returns
    -------
    F-value : float
        The computed F-value of the test.
    p-value : float
        The associated p-value from the F-distribution.
    rankings : array_like
        The ranking for each group.
    pivots : array_like
        The pivotal quantities for each group.

    References
    ----------
    M. Friedman, The use of ranks to avoid the assumption of normality implicit in the analysis of variance, Journal of the American Statistical Association 32 (1937) 674–701.
    D.J. Sheskin, Handbook of parametric and nonparametric statistical procedures. crc Press, 2003, Test 25: The Friedman Two-Way Analysis of Variance by Ranks
    """
    k = len(args)
    if k < 2:
        raise ValueError("Less than 2 levels")
    n = len(args[0])
    if len(set([len(v) for v in args])) != 1:
        raise ValueError("Unequal number of samples")

    rankings = []
    for i in range(n):
        row = [col[i] for col in args]
        row_sort = sorted(row)
        rankings.append(
            [row_sort.index(v) + 1 + (row_sort.count(v) - 1) / 2.0 for v in row]
        )

    rankings_avg = [np.mean([case[j] for case in rankings]) for j in range(k)]
    rankings_cmp = [
        r / np.lib.scimath.sqrt(k * (k + 1) / (6.0 * n)) for r in rankings_avg
    ]

    chi2 = ((12 * n) / float((k * (k + 1)))) * (
        (sum(r ** 2 for r in rankings_avg)) - ((k * (k + 1) ** 2) / float(4))
    )
    iman_davenport = ((n - 1) * chi2) / float((n * (k - 1) - chi2))

    p_value = 1 - sp.stats.f.cdf(iman_davenport, k - 1, (k - 1) * (n - 1))

    return iman_davenport, p_value, rankings_avg, rankings_cmp


def bonferroni_dunn_test(ranks: dict, control: str = None) -> [list, list, list, list]:
    """
    Performs a Bonferroni-Dunn post-hoc test using the pivot quantities obtained by a ranking test.
    Tests the hypothesis that the ranking of the control method is different to each of the other methods.

    Parameters
    ----------
    ranks : dictionary_like
        A dictionary with format 'groupname':'pivotal quantity'
    control : string optional
        The name of the control method (one vs all), default None (all vs all)

    Returns
    ----------
    comparisons : array-like
        Strings identifier of each comparison with format 'group_i vs group_j'
    Z-values : array-like
        The computed Z-value statistic for each comparison.
    p-values : array-like
        The associated p-value from the Z-distribution which depends on the index of the comparison
    Adjusted p-values : array-like
        The associated adjusted p-values which can be compared with a significance level

    References
    ----------
    O.J. Dunn, Multiple comparisons among means, Journal of the American Statistical Association 56 (1961) 52–64.
    """
    k = len(ranks)
    values = list(ranks.values())
    keys = list(ranks.keys())

    if control is None:
        control_i = values.index(min(values))
    else:
        control_i = keys.index(control)

    comparisons = [
        keys[control_i] + " vs " + keys[i] for i in range(k) if i != control_i
    ]
    z_values = [abs(values[control_i] - values[i]) for i in range(k) if i != control_i]
    p_values = [2 * (1 - sp.stats.norm.cdf(abs(z))) for z in z_values]
    # Sort values by p_value so that p_0 < p_1
    p_values, z_values, comparisons = map(
        list, zip(*sorted(zip(p_values, z_values, comparisons), key=lambda t: t[0]))
    )
    adj_p_values = [min((k - 1) * p_value, 1) for p_value in p_values]

    return comparisons, z_values, p_values, adj_p_values



import numpy as np


# Step 1: normalize the decision matrix
def norm(x, y):
    """normalization function; x is the array with the
    performances and y is the normalization method.
    For vector input 'v' and for linear 'l'
    """

    if y == "v":
        k = np.array(np.cumsum(x ** 2, 0))
        z = np.array(
            [
                [
                    round(x[i, j] / np.sqrt(k[x.shape[0] - 1, j]), 3)
                    for j in range(x.shape[1])
                ]
                for i in range(x.shape[0])
            ]
        )
        return z
    else:
        yy = []
        k = []
        for i in range(x.shape[1]):
            yy.append(np.amax(x[:, i : i + 1]))
            k = np.array(yy)

        z = np.array(
            [
                [round(x[i, j] / k[j], 3) for j in range(x.shape[1])]
                for i in range(x.shape[0])
            ]
        )
        return z


# Step 2: find the weighted normalized decision matrix
def mul_w(r, t):
    """multiplication of each evaluation by the associate
    weight; r stands for the weights matrix and t for
    the normalized matrix resulting from norm()
    """
    z = np.array(
        [
            [round(t[i, j] * r[j], 3) for j in range(t.shape[1])]
            for i in range(t.shape[0])
        ]
    )
    return z


# Step 3: calculate the ideal and anti-ideal solutions
def zenith_nadir(x, y):
    """zenith and nadir virtual action function; x is the
    weighted normalized decision matrix and y is the
    action used. For min/max input 'm' and for absolute
    input enter 'a'
    """
    if y == "m":
        bb = []
        cc = []
        for i in range(x.shape[1]):
            bb.append(np.amax(x[:, i : i + 1]))
            b = np.array(bb)
            cc.append(np.amin(x[:, i : i + 1]))
            c = np.array(cc)
        return b, c
    else:
        b = np.ones(x.shape[1])
        c = np.zeros(x.shape[1])
        return b, c


# Step 4: determine the distance to the ideal and anti-ideal
# solutions
def distance(x, y, z):
    """calculate the distances to the ideal solution (di+)
    and the anti-ideal solution (di-); x is the result
    of mul_w() and y, z the results of zenith_nadir()
    """
    a = np.array(
        [[(x[i, j] - y[j]) ** 2 for j in range(x.shape[1])] for i in range(x.shape[0])]
    )
    b = np.array(
        [[(x[i, j] - z[j]) ** 2 for j in range(x.shape[1])] for i in range(x.shape[0])]
    )
    return np.sqrt(sum(a, 1)), np.sqrt(sum(b, 1))


# TOPSIS method: it calls the other functions and includes
# step 5
def topsis(matrix, weight, norm_m, id_sol):
    """matrix is the initial decision matrix, weight is
    the weights matrix, norm_m is the normalization
    method, id_sol is the action used, and pl is 'y'
    for plotting the results or any other string for
    not
    """
    z = mul_w(weight, norm(matrix, norm_m))
    s, f = zenith_nadir(z, id_sol)
    p, n = distance(z, s, f)
    final_s = np.array([n[i] / (p[i] + n[i]) for i in range(p.shape[0])])

    return final_s



from setuptools import setup, find_packages
from codecs import open
from os import path

__author__ = 'Giulio Rossetti'
__license__ = "BSD-2-Clause"
__email__ = "giulio.rossetti@gmail.com"


here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    requirements = f.read().splitlines()


setup(name='cdlib',
      version='0.2.6',
      license='BSD-Clause-2',
      description='Community Discovery Library',
      url='https://github.com/GiulioRossetti/cdlib',
      author='Giulio Rossetti',
      author_email='giulio.rossetti@gmail.com',
      classifiers=[
          # How mature is this project? Common values are
          #   3 - Alpha
          #   4 - Beta
          #   5 - Production/Stable
          'Development Status :: 4 - Beta',

          # Indicate who your project is intended for
          'Intended Audience :: Developers',
          'Topic :: Software Development :: Build Tools',

          # Pick your license as you wish (should match "license" above)
          'License :: OSI Approved :: BSD License',

          "Operating System :: POSIX :: Other",
          "Operating System :: MacOS",

          # Specify the Python versions you support here. In particular, ensure
          # that you indicate whether you support Python 2, Python 3 or both.
          'Programming Language :: Python',
          'Programming Language :: Python :: 3'
      ],
      keywords='community-discovery node-clustering edge-clustering complex-networks',
      install_requires=requirements,
      long_description=long_description,
      long_description_content_type='text/markdown',
      extras_require={
        'C':  ["infomap>=1.3.0", "wurlitzer>=1.0.2", "GraphRicciCurvature", "networkit", "pycombo","leidenalg","karateclub"],
      },
      packages=find_packages(exclude=["*.test", "*.test.*", "test.*", "test", "cdlib.test", "cdlib.test.*"]),
      )



import numpy as np
from cdlib.evaluation.internal import onmi
from cdlib.evaluation.internal.omega import Omega
from nf1 import NF1
from collections import namedtuple, defaultdict

__all__ = [
    "MatchingResult",
    "normalized_mutual_information",
    "overlapping_normalized_mutual_information_LFK",
    "overlapping_normalized_mutual_information_MGH",
    "omega",
    "f1",
    "nf1",
    "adjusted_rand_index",
    "adjusted_mutual_information",
    "variation_of_information",
    "partition_closeness_simple",
]

# MatchingResult = namedtuple("MatchingResult", ['mean', 'std'])

MatchingResult = namedtuple("MatchingResult", "score std")
MatchingResult.__new__.__defaults__ = (None,) * len(MatchingResult._fields)


def __check_partition_coverage(first_partition: object, second_partition: object):
    nodes_first = {
        node: None for community in first_partition.communities for node in community
    }
    nodes_second = {
        node: None for community in second_partition.communities for node in community
    }

    if len(set(nodes_first.keys()) ^ set(nodes_second.keys())) != 0:
        raise ValueError("Both partitions should cover the same node set")


def __check_partition_overlap(first_partition: object, second_partition: object):
    if first_partition.overlap or second_partition.overlap:
        raise ValueError("Not defined for overlapping partitions")


def normalized_mutual_information(
    first_partition: object, second_partition: object
) -> MatchingResult:
    """
    Normalized Mutual Information between two clusterings.

    Normalized Mutual Information (NMI) is an normalization of the Mutual
    Information (MI) score to scale the results between 0 (no mutual
    information) and 1 (perfect correlation). In this function, mutual
    information is normalized by ``sqrt(H(labels_true) * H(labels_pred))``

    :param first_partition: NodeClustering object
    :param second_partition: NodeClustering object
    :return: MatchingResult object

    :Example:

      >>> from cdlib import evaluation, algorithms
      >>> g = nx.karate_club_graph()
      >>> louvain_communities = algorithms.louvain(g)
      >>> leiden_communities = algorithms.leiden(g)
      >>> evaluation.normalized_mutual_information(louvain_communities,leiden_communities)

    """

    __check_partition_coverage(first_partition, second_partition)
    __check_partition_overlap(first_partition, second_partition)

    first_partition_c = [
        x[1]
        for x in sorted(
            [
                (node, nid)
                for nid, cluster in enumerate(first_partition.communities)
                for node in cluster
            ],
            key=lambda x: x[0],
        )
    ]

    second_partition_c = [
        x[1]
        for x in sorted(
            [
                (node, nid)
                for nid, cluster in enumerate(second_partition.communities)
                for node in cluster
            ],
            key=lambda x: x[0],
        )
    ]

    from sklearn.metrics import normalized_mutual_info_score

    return MatchingResult(
        score=normalized_mutual_info_score(first_partition_c, second_partition_c)
    )


def overlapping_normalized_mutual_information_LFK(
    first_partition: object, second_partition: object
) -> MatchingResult:
    """
    Overlapping Normalized Mutual Information between two clusterings.

    Extension of the Normalized Mutual Information (NMI) score to cope with overlapping partitions.
    This is the version proposed by Lancichinetti et al. (1)



    :param first_partition: NodeClustering object
    :param second_partition: NodeClustering object
    :return: MatchingResult object

    :Example:

    >>> from cdlib import evaluation, algorithms
    >>> g = nx.karate_club_graph()
    >>> louvain_communities = algorithms.louvain(g)
    >>> leiden_communities = algorithms.leiden(g)
    >>> evaluation.overlapping_normalized_mutual_information_LFK(louvain_communities,leiden_communities)
    :Reference:

    1. Lancichinetti, A., Fortunato, S., & Kertesz, J. (2009). Detecting the overlapping and hierarchical community structure in complex networks. New Journal of Physics, 11(3), 033015.
    """

    return MatchingResult(
        score=onmi.onmi(
            [set(x) for x in first_partition.communities],
            [set(x) for x in second_partition.communities],
        )
    )


def overlapping_normalized_mutual_information_MGH(
    first_partition: object, second_partition: object, normalization: str = "max"
) -> MatchingResult:
    """
    Overlapping Normalized Mutual Information between two clusterings.

    Extension of the Normalized Mutual Information (NMI) score to cope with overlapping partitions.
    This is the version proposed by McDaid et al. using a different normalization than the original LFR one. See ref.
    for more details.

    :param first_partition: NodeClustering object
    :param second_partition: NodeClustering object
    :param normalization: one of "max" or "LFK". Default "max" (corresponds to the main method described in the article)
    :return: MatchingResult object

    :Example:

    >>> from cdlib import evaluation, algorithms
    >>> g = nx.karate_club_graph()
    >>> louvain_communities = algorithms.louvain(g)
    >>> leiden_communities = algorithms.leiden(g)
    >>> evaluation.overlapping_normalized_mutual_information_MGH(louvain_communities,leiden_communities)
    :Reference:

    1. McDaid, A. F., Greene, D., & Hurley, N. (2011). Normalized mutual information to evaluate overlapping community finding algorithms. arXiv preprint arXiv:1110.2515. Chicago
    """

    if normalization == "max":
        variant = "MGH"
    elif normalization == "LFK":
        variant = "MGH_LFK"
    else:
        raise ValueError(
            "Wrong 'normalization' value. Please specify one among [max, LFK]."
        )

    return MatchingResult(
        score=onmi.onmi(
            [set(x) for x in first_partition.communities],
            [set(x) for x in second_partition.communities],
            variant=variant,
        )
    )


def omega(first_partition: object, second_partition: object) -> MatchingResult:
    """
    Index of resemblance for overlapping, complete coverage, network clusterings.

    :param first_partition: NodeClustering object
    :param second_partition: NodeClustering object
    :return: MatchingResult object

    :Example:

    >>> from cdlib import evaluation, algorithms
    >>> g = nx.karate_club_graph()
    >>> louvain_communities = algorithms.louvain(g)
    >>> leiden_communities = algorithms.leiden(g)
    >>> evaluation.omega(louvain_communities,leiden_communities)
    :Reference:

    1. Gabriel Murray, Giuseppe Carenini, and Raymond Ng. 2012. `Using the omega index for evaluating abstractive algorithms detection. <https://pdfs.semanticscholar.org/59d6/5d5aa09d789408fd9fd3c009a1b070ff5859.pdf/>`_ In Proceedings of Workshop on Evaluation Metrics and System Comparison for Automatic Summarization. Association for Computational Linguistics, Stroudsburg, PA, USA, 10-18.
    """

    __check_partition_coverage(first_partition, second_partition)

    first_partition = {k: v for k, v in enumerate(first_partition.communities)}
    second_partition = {k: v for k, v in enumerate(second_partition.communities)}

    om_idx = Omega(first_partition, second_partition)
    return MatchingResult(score=om_idx.omega_score)


def f1(first_partition: object, second_partition: object) -> MatchingResult:
    """
    Compute the average F1 score of the optimal algorithms matches among the partitions in input.
    Works on overlapping/non-overlapping complete/partial coverage partitions.

    :param first_partition: NodeClustering object
    :param second_partition: NodeClustering object
    :return: MatchingResult object

    :Example:

    >>> from cdlib import evaluation, algorithms
    >>> g = nx.karate_club_graph()
    >>> louvain_communities = algorithms.louvain(g)
    >>> leiden_communities = algorithms.leiden(g)
    >>> evaluation.f1(louvain_communities,leiden_communities)

    :Reference:

    1. Rossetti, G., Pappalardo, L., & Rinzivillo, S. (2016). `A novel approach to evaluate algorithms detection internal on ground truth. <https://www.researchgate.net/publication/287204505_A_novel_approach_to_evaluate_community_detection_algorithms_on_ground_truth/>`_ In Complex Networks VII (pp. 133-144). Springer, Cham.
    """

    nf = NF1(first_partition.communities, second_partition.communities)
    results = nf.summary()
    return MatchingResult(
        score=results["details"]["F1 mean"][0], std=results["details"]["F1 std"][0]
    )


def nf1(first_partition: object, second_partition: object) -> MatchingResult:
    """
    Compute the Normalized F1 score of the optimal algorithms matches among the partitions in input.
    Works on overlapping/non-overlapping complete/partial coverage partitions.

    :param first_partition: NodeClustering object
    :param second_partition: NodeClustering object
    :return: MatchingResult object

    :Example:

    >>> from cdlib import evaluation, algorithms
    >>> g = nx.karate_club_graph()
    >>> louvain_communities = algorithms.louvain(g)
    >>> leiden_communities = algorithms.leiden(g)
    >>> evaluation.nf1(louvain_communities,leiden_communities)

    :Reference:

    1. Rossetti, G., Pappalardo, L., & Rinzivillo, S. (2016). `A novel approach to evaluate algorithms detection internal on ground truth. <https://www.researchgate.net/publication/287204505_A_novel_approach_to_evaluate_community_detection_algorithms_on_ground_truth/>`_

    2. Rossetti, G. (2017). : `RDyn: graph benchmark handling algorithms dynamics. Journal of Complex Networks. <https://academic.oup.com/comnet/article-abstract/5/6/893/3925036?redirectedFrom=PDF/>`_ 5(6), 893-912.
    """

    nf = NF1(first_partition.communities, second_partition.communities)
    results = nf.summary()
    return MatchingResult(score=results["scores"].loc["NF1"][0])


def adjusted_rand_index(
    first_partition: object, second_partition: object
) -> MatchingResult:
    """Rand index adjusted for chance.

    The Rand Index computes a similarity measure between two clusterings
    by considering all pairs of samples and counting pairs that are
    assigned in the same or different clusters in the predicted and
    true clusterings.

    The raw RI score is then "adjusted for chance" into the ARI score
    using the following scheme::

        ARI = (RI - Expected_RI) / (max(RI) - Expected_RI)

    The adjusted Rand index is thus ensured to have a value close to
    0.0 for random labeling independently of the number of clusters and
    samples and exactly 1.0 when the clusterings are identical (up to
    a permutation).

    ARI is a symmetric measure::

        adjusted_rand_index(a, b) == adjusted_rand_index(b, a)

    :param first_partition: NodeClustering object
    :param second_partition: NodeClustering object
    :return: MatchingResult object

    :Example:

    >>> from cdlib import evaluation, algorithms
    >>> g = nx.karate_club_graph()
    >>> louvain_communities = algorithms.louvain(g)
    >>> leiden_communities = algorithms.leiden(g)
    >>> evaluation.adjusted_rand_index(louvain_communities,leiden_communities)

    :Reference:

    1. Hubert, L., & Arabie, P. (1985). `Comparing partitions. <https://link.springer.com/article/10.1007/BF01908075/>`_ Journal of classification, 2(1), 193-218.
    """

    __check_partition_coverage(first_partition, second_partition)
    __check_partition_overlap(first_partition, second_partition)

    first_partition_c = [
        x[1]
        for x in sorted(
            [
                (node, nid)
                for nid, cluster in enumerate(first_partition.communities)
                for node in cluster
            ],
            key=lambda x: x[0],
        )
    ]

    second_partition_c = [
        x[1]
        for x in sorted(
            [
                (node, nid)
                for nid, cluster in enumerate(second_partition.communities)
                for node in cluster
            ],
            key=lambda x: x[0],
        )
    ]

    from sklearn.metrics import adjusted_rand_score

    return MatchingResult(
        score=adjusted_rand_score(first_partition_c, second_partition_c)
    )


def adjusted_mutual_information(
    first_partition: object, second_partition: object
) -> MatchingResult:
    """Adjusted Mutual Information between two clusterings.

    Adjusted Mutual Information (AMI) is an adjustment of the Mutual
    Information (MI) score to account for chance. It accounts for the fact that
    the MI is generally higher for two clusterings with a larger number of
    clusters, regardless of whether there is actually more information shared.
    For two clusterings :math:`U` and :math:`V`, the AMI is given as::

        AMI(U, V) = [MI(U, V) - E(MI(U, V))] / [max(H(U), H(V)) - E(MI(U, V))]

    This metric is independent of the absolute values of the labels:
    a permutation of the class or cluster label values won't change the
    score value in any way.

    This metric is furthermore symmetric: switching ``label_true`` with
    ``label_pred`` will return the same score value. This can be useful to
    measure the agreement of two independent label assignments strategies
    on the same dataset when the real ground truth is not known.

    Be mindful that this function is an order of magnitude slower than other
    metrics, such as the Adjusted Rand Index.

    :param first_partition: NodeClustering object
    :param second_partition: NodeClustering object
    :return: MatchingResult object

    :Example:

    >>> from cdlib import evaluation, algorithms
    >>> g = nx.karate_club_graph()
    >>> louvain_communities = algorithms.louvain(g)
    >>> leiden_communities = algorithms.leiden(g)
    >>> evaluation.adjusted_mutual_information(louvain_communities,leiden_communities)

    :Reference:

    1. Vinh, N. X., Epps, J., & Bailey, J. (2010). `Information theoretic measures for clusterings comparison: Variants, properties, normalization and correction for chance. <http://jmlr.csail.mit.edu/papers/volume11/vinh10a/vinh10a.pdf/>`_ Journal of Machine Learning Research, 11(Oct), 2837-2854.
    """

    __check_partition_coverage(first_partition, second_partition)
    __check_partition_overlap(first_partition, second_partition)

    first_partition_c = [
        x[1]
        for x in sorted(
            [
                (node, nid)
                for nid, cluster in enumerate(first_partition.communities)
                for node in cluster
            ],
            key=lambda x: x[0],
        )
    ]

    second_partition_c = [
        x[1]
        for x in sorted(
            [
                (node, nid)
                for nid, cluster in enumerate(second_partition.communities)
                for node in cluster
            ],
            key=lambda x: x[0],
        )
    ]

    from sklearn.metrics import adjusted_mutual_info_score

    return MatchingResult(
        score=adjusted_mutual_info_score(first_partition_c, second_partition_c)
    )


def variation_of_information(
    first_partition: object, second_partition: object
) -> MatchingResult:
    """Variation of Information among two nodes partitions.

    $$ H(p)+H(q)-2MI(p, q) $$

    where MI is the mutual information, H the partition entropy and p,q are the algorithms sets

    :param first_partition: NodeClustering object
    :param second_partition: NodeClustering object
    :return: MatchingResult object

    :Example:

    >>> from cdlib import evaluation, algorithms
    >>> g = nx.karate_club_graph()
    >>> louvain_communities = algorithms.louvain(g)
    >>> leiden_communities = algorithms.leiden(g)
    >>> evaluation.variation_of_information(louvain_communities,leiden_communities)

    :Reference:

    1. Meila, M. (2007). `Comparing clusterings - an information based distance. <https://www.sciencedirect.com/science/article/pii/S0047259X06002016/>`_ Journal of Multivariate Analysis, 98, 873-895. doi:10.1016/j.jmva.2006.11.013
    """

    __check_partition_coverage(first_partition, second_partition)
    __check_partition_overlap(first_partition, second_partition)

    n = float(sum([len(c1) for c1 in first_partition.communities]))
    sigma = 0.0
    for c1 in first_partition.communities:
        p = len(c1) / n
        for c2 in second_partition.communities:
            q = len(c2) / n
            r = len(set(c1) & set(c2)) / n
            if r > 0.0:
                sigma += r * (np.log2(r / p) + np.log2(r / q))

    return MatchingResult(score=abs(sigma))


def partition_closeness_simple(
    first_partition: object, second_partition: object
) -> MatchingResult:
    """Community size density closeness.
    Simple implementation that does not leverage kernel density estimator.

    $$ S_G(A,B) = \frac{1}{2} \Sum_{i=1}^{r}\Sum_{j=1}^{s} min(\frac{n^a(x^a_i)}{N^a}, \frac{n^b_j(x^b_j)}{N^b}) \delta(x_i^a,x_j^b) $$

    where:

    $$ N^a $$ total number of communities in A of any size;
    $$ x^a $$ ordered list of community sizes for A;
    $$ n^a $$ multiplicity of community sizes for A.

    (symmetrically for B)

    :param first_partition: NodeClustering object
    :param second_partition: NodeClustering object
    :return: MatchingResult object

    :Example:

    >>> from cdlib import evaluation, algorithms
    >>> g = nx.karate_club_graph()
    >>> louvain_communities = algorithms.louvain(g)
    >>> leiden_communities = algorithms.leiden(g)
    >>> evaluation.partition_closeness_simple(louvain_communities,leiden_communities)

    :Reference:

    1. Dao, Vinh-Loc, Cécile Bothorel, and Philippe Lenca. "Estimating the similarity of community detection methods based on cluster size distribution." International Conference on Complex Networks and their Applications. Springer, Cham, 2018.
    """
    coms_a = sorted(list(set([len(c) for c in first_partition.communities])))
    freq_a = defaultdict(int)
    for a in coms_a:
        freq_a[a] += 1
    freq_a = [freq_a[a] for a in sorted(freq_a)]
    n_a = sum([coms_a[i] * freq_a[i] for i in range(0, len(coms_a))])

    coms_b = sorted(list(set([len(c) for c in second_partition.communities])))
    freq_b = defaultdict(int)
    for b in coms_b:
        freq_b[b] += 1
    freq_b = [freq_b[b] for b in sorted(freq_b)]
    n_b = sum([coms_b[i] * freq_b[i] for i in range(0, len(coms_b))])

    closeness = 0
    for i in range(0, len(coms_a)):
        for j in range(0, len(coms_b)):
            if coms_a[i] == coms_b[j]:
                closeness += min(
                    (coms_a[i] * freq_a[i]) / n_a, (coms_b[j] * freq_b[j]) / n_b
                )
    closeness *= 0.5

    return MatchingResult(score=closeness)



"""
实现LPANNI [1]。
[1] Meilian L , Zhenlin Z , Zhihe Q , et al.
LPANNI: Overlapping Community Detection Using Label Propagation in Large-Scale Complex Networks[J].
IEEE Transactions on Knowledge and Data Engineering, 2018, PP:1-1.
"""

import networkx as nx


def compute_NI(g: nx.Graph):
    """
    计算网络中每个节点的ni，将结果存储在节点的NI属性中
    :param g: 网络
    :return:None
    """
    max_ni = 0
    min_ni = 2 * g.number_of_nodes()
    for node in g.nodes:
        neighbors = list(g.adj[node].keys())
        triangle_num = 0
        for i in neighbors:
            for j in neighbors:
                if i < j and g.has_edge(i, j):
                    triangle_num += 1
        ni = g.degree[node] + triangle_num
        g.nodes[node]["NI"] = ni
        max_ni = max(max_ni, ni)
        min_ni = min(min_ni, ni)

    for node in g.nodes:
        g.nodes[node]["NI"] = 0.5 + 0.5 * (g.nodes[node]["NI"] - min_ni) / (
            max_ni - min_ni
        )


def compute_sim(g: nx.Graph):
    """
    计算网络中节点u,v的相似度，其中(u,v)∈E
    根据[1]文的考量，将α设置为3.
    (u,v)∈E是因为在LPA算法中只有相邻节点才会互相传播标签
    不存在边的那些节点对计算了也没用
    :param g:网络图
    :return:None
    """
    for edge in g.edges:
        u, v = edge[0], edge[1]
        # p=1
        s = 1.0
        # 开始计算p=2的情况
        for i in list(g.adj[u].keys()):
            if g.has_edge(i, v):
                s += 1 / 2
        # 开始计算p=3的情况
        for i in list(g.adj[u].keys()):
            for j in list(g.adj[i].keys()):
                if g.has_edge(j, v) and (not j == u):
                    s += 1 / 3
        for i in (u, v):
            if g.nodes[i].get("s", -1) == -1:
                g.nodes[i]["s"] = {}
        g.nodes[u]["s"][v] = s
        g.nodes[v]["s"][u] = s

    for edge in g.edges:
        u, v = edge[0], edge[1]
        for i in (u, v):
            if g.nodes[i].get("sim", -1) == -1:
                g.nodes[i]["sim"] = {}
        sim = (
            g.nodes[u]["s"][v]
            / (sum(g.nodes[u]["s"].values()) * sum(g.nodes[v]["s"].values())) ** 0.5
        )
        g.nodes[u]["sim"][v] = sim
        g.nodes[v]["sim"][u] = sim


def compute_NNI(g: nx.Graph):
    """
    计算节点v对节点u的影响力，其中(v,u)∈E
    u节点的属性'NNI'是一个字典，其中包含所有与u相邻的节点V
    对于每一个v∈V，g.nodes[u]['NNI'][v]表示v给u造成的影响力
    :param g:网络图
    :return:None
    """
    for u in g.nodes:
        g.nodes[u]["NNI"] = {}
        sim_max = max(g.nodes[u]["sim"].values())
        for v in list(g.adj[u].keys()):
            g.nodes[u]["NNI"][v] = (
                g.nodes[v]["NI"] * g.nodes[u]["sim"][v] / sim_max
            ) ** 0.5


def LPANNI(g: nx.Graph):
    """
    实现LPANNI对网络进行社区划分
    :param g: 网络图
    :return: None
    """
    compute_NI(g)
    compute_sim(g)
    compute_NNI(g)

    nodes = list(g.nodes)
    v_queue = []
    for node in nodes:
        v_queue.append((node, g.nodes[node]["NI"]))
        g.nodes[node]["L"] = {node: 1}
        g.nodes[node]["dominant"] = 1
        g.nodes[node]["label"] = node
    v_queue = sorted(v_queue, key=lambda v: v[1])
    nodes = [v[0] for v in v_queue]

    # 定义最大迭代次数
    T = 10
    t = 0
    while t < T:
        change = False
        for node in nodes:
            L_Ng = {}
            # 计算邻居们的标签和权重
            for neighbor in list(g.adj[node].keys()):
                c, b = (
                    g.nodes[neighbor]["label"],
                    g.nodes[neighbor]["dominant"] * g.nodes[node]["NNI"][neighbor],
                )
                if L_Ng.get(c, -1) == -1:
                    L_Ng[c] = b
                else:
                    L_Ng[c] += b
            # 除去权重过小的标签
            avg = sum(L_Ng.values()) / len(L_Ng)
            max_dominant = 0
            label = -1
            g.nodes[node]["L"] = {}
            for c in L_Ng.keys():
                if L_Ng[c] >= avg:
                    g.nodes[node]["L"][c] = L_Ng[c]
                    if L_Ng[c] > max_dominant:
                        max_dominant = L_Ng[c]
                        label = c
            sum_dominant = sum(g.nodes[node]["L"].values())

            for c in g.nodes[node]["L"].keys():
                g.nodes[node]["L"][c] /= sum_dominant

            if not g.nodes[node]["label"] == label:
                g.nodes[node]["label"] = label
                change = True
            g.nodes[node]["dominant"] = max_dominant / sum_dominant
        if not change:
            break
        t += 1


class GraphGenerator:
    """
    定义归属系数的阈值
    也就是说，只有一个节点对一个社区的归属系数大于这个阈值时，我们才考虑将这个节点加入这个社区中
    """

    b_threshold = 0.0
    g = nx.Graph()

    def __init__(self, threshold, g):
        self.b_threshold = threshold
        self.g = g

    def get_Overlapping_communities(self) -> list:
        """
        从图中将所有的重叠社区返回
        :return: list
        """
        d = {}
        for node in self.g.nodes:
            L = self.g.nodes[node]["L"]
            for label in L.keys():
                if L[label] > self.b_threshold:
                    # 这个节点属于label 社区
                    if d.get(label, -1) == -1:
                        d[label] = {node}
                    else:
                        d[label].add(node)
        return list(d.values())

    def get_Overlapping_nodes(self) -> set:
        """
        从图中将所有的重叠节点返回
        :return: 所有的重叠节点
        """
        overlapping_nodes = set()
        for node in self.g.nodes:
            L = self.g.nodes[node]["L"]
            count = 0
            for label in L.keys():
                if L[label] > self.b_threshold:
                    count += 1
                    if count >= 2:
                        overlapping_nodes.add(node)
                        break
        return overlapping_nodes



# Author: True Price <jtprice@cs.unc.edu>
from itertools import combinations
from collections import defaultdict


def dp_clus(g, weights=None, d_threshold=0.9, cp_threshold=0.5, overlap=True):
    if weights is None:
        if overlap:
            return __dpclus(g, d_threshold, cp_threshold)
        else:
            return __dpclus_crisp(g, d_threshold, cp_threshold)
    else:
        return __dpclus_w(g, d_threshold, cp_threshold, weights)


def __dpclus(g, d_threshold=0.9, cp_threshold=0.5):
    res = []
    data = defaultdict(set)  # node id => neighboring node ids

    # to match the original DPClus output, we need to keep track of the order in
    # which nodes were read from the file, in order to maintain cluster order when
    # breaking ties in our priority ranking (values are decreasing, i.e. negative)
    node_index = dict()
    b_list = []  # the original algorithm indexes column b after column a... >:-(

    # read in graph
    for a, b in g.edges():
        data[a].add(b)
        data[b].add(a)
        node_index.setdefault(a, -len(node_index))
        # node_index.setdefault(b, -len(node_index))
        b_list.append(b)

    for b in b_list:
        node_index.setdefault(b, -len(node_index))

    unvisited = set(data)
    num_clusters = 0

    while unvisited:
        # get highest degree node
        seed = max(unvisited, key=lambda k: (len(data[k] & unvisited), node_index[k]))
        frontier = data[seed] & unvisited
        if not frontier:
            break  # no connections left to analyze

        edges, weights = defaultdict(lambda: defaultdict(int)), defaultdict(int)
        for a, b in combinations(unvisited, 2):
            if b not in data[a]:
                continue
            shared = len(data[a] & data[b] & unvisited)
            edges[a][b], edges[b][a] = shared, shared
            weights[a] += shared
            weights[b] += shared

        max_w, _, node = max((w, node_index[n], n) for n, w in weights.items())
        if max_w > 0:
            seed = node
            frontier = data[seed] & unvisited

        cluster = {seed}
        cluster_degrees = {seed: 0}
        nn, ne = 1, 0  # number of nodes, edges in cluster

        while frontier:
            # find higest priority node:
            # 0. number of edges between node and cluster nodes
            # 1. sum of edge weights between node and cluster nodes
            # 2. the node's index
            # 3. the node itself
            e_nk, _, _, p = max(
                (
                    len(data[n] & cluster),
                    sum(edges[n][c] for c in cluster),
                    node_index[n],
                    n,
                )
                for n in frontier
            )

            density = 2.0 * (ne + e_nk) / (nn * (nn + 1))
            if density < d_threshold:
                break  # adding the node gives too low density; cluster is finished

            # if all nodes only have one connecting edge in cluster, use "fine-tuning"
            # this orders by the number of neighbors in the frontier, minus the
            # connectedness of the attached (cluster) node within the cluster
            cp = cp_threshold
            if e_nk == 1 and len(cluster) > 1:
                n_degree = dict()  # node => fine-tuning parameter
                for n in frontier:
                    for c in cluster:  # find adjacent cluster node
                        if n in edges[c]:
                            break
                    n_degree[n] = len(data[n] & frontier) - cluster_degrees[c]
                p = max(frontier, key=lambda k: (n_degree[k], node_index[k]))
                if n_degree[p] > 0:
                    cp /= 2.0
            if (e_nk / density / (nn + 1)) < cp:
                break  # no good node found; cluster is finished

            # otherwise, add the node to the cluster
            cluster.add(p)
            nn, ne = (nn + 1), (ne + e_nk)

            cluster_degrees[p] = e_nk
            for n in data[p] & cluster:
                cluster_degrees[n] += 1

            frontier = set.union(*((data[n] - cluster) & unvisited for n in cluster))

        # add overlapping nodes
        # frontier[2] stores our fine-tuning parameter, in this case
        frontier_nodes = set.union(*(data[c] for c in cluster)) - cluster
        frontier = sorted(
            [
                len(data[n] & cluster),
                sum(edges[n][c] for c in cluster),
                0,
                node_index[n],
                n,
            ]
            for n in frontier_nodes
        )
        # "fine-tuning"
        fine_tuning = False
        if frontier and frontier[-1][0] == 1:
            for n in frontier:
                for c in cluster:  # find adjacent cluster node
                    if n[4] in edges[c]:
                        break
                n[2] = len(data[n[4]] & frontier_nodes) - cluster_degrees[c]
            frontier.sort(key=lambda k: (k[2], k[3]))
            fine_tuning = True
        # iterate through visited neighbors and update accordingly
        while frontier:
            e_nk, _, w, _, p = frontier.pop()
            cp = cp_threshold / 2.0 if fine_tuning and w > 0 else cp_threshold

            density = 2.0 * (ne + e_nk) / (nn * (nn + 1))
            if density < d_threshold or (e_nk / density / (nn + 1)) < cp:
                continue

            # add node to the cluster
            cluster.add(p)
            nn, ne = (nn + 1), (ne + e_nk)

            cluster_degrees[p] = e_nk
            for n in data[p] & cluster:
                cluster_degrees[n] += 1

            # update E_nk for the other nodes on the frontier
            for n in frontier:
                if p in edges[n[4]]:
                    n[0] += 1

        unvisited -= cluster
        res.append(list(cluster))

        num_clusters += 1

    return res


def __dpclus_crisp(g, d_threshold, cp_threshold):
    res = []
    data = defaultdict(set)  # node id => neighboring node ids

    # to match the original DPClus output, we need to keep track of the order in
    # which nodes were read from the file, in order to maintain cluster order when
    # breaking ties in our priority ranking (values are decreasing, i.e. negative)
    node_index = dict()
    b_list = []  # the original algorithm indexes column b after column a... >:-(

    # read in graph
    for a, b in g.edges():
        data[a].add(b)
        data[b].add(a)
        node_index.setdefault(a, -len(node_index))
        # node_index.setdefault(b, -len(node_index))
        b_list.append(b)
    for b in b_list:
        node_index.setdefault(b, -len(node_index))

    unvisited = set(data)
    num_clusters = 0

    while unvisited:
        # get highest degree node
        seed = max(unvisited, key=lambda k: (len(data[k] & unvisited), node_index[k]))
        frontier = data[seed] & unvisited
        if not frontier:
            break  # no connections left to analyze

        edges, weights = defaultdict(lambda: defaultdict(int)), defaultdict(int)
        for a, b in combinations(unvisited, 2):
            if b not in data[a]:
                continue
            shared = len(data[a] & data[b] & unvisited)
            edges[a][b], edges[b][a] = shared, shared
            weights[a] += shared
            weights[b] += shared

        max_w, _, node = max((w, node_index[n], n) for n, w in weights.items())
        if max_w > 0:
            seed = node
            frontier = data[seed] & unvisited

        cluster = {seed}
        cluster_degrees = {seed: 0}
        nn, ne = 1, 0  # number of nodes, edges in cluster

        while frontier:
            # find higest priority node:
            # 0. number of edges between node and cluster nodes
            # 1. sum of edge weights between node and cluster nodes
            # 2. the node itself, as indexed by node_index
            e_nk, _, _, p = max(
                (
                    len(data[n] & cluster),
                    sum(edges[n][c] for c in cluster),
                    node_index[n],
                    n,
                )
                for n in frontier
            )

            density = 2.0 * (ne + e_nk) / (nn * (nn + 1))
            if density < d_threshold:
                break  # adding the node gives too low density; cluster is finished

            # if all nodes only have one connecting edge in cluster, use "fine-tuning"
            # this orders by the number of neighbors in the frontier, minus the
            # connectedness of the attached (cluster) node within the cluster
            cp = cp_threshold
            if e_nk == 1 and len(cluster) > 1:
                n_degree = dict()  # node => fine-tuning parameter
                for n in frontier:
                    for c in cluster:  # find adjacent cluster node
                        if n in edges[c]:
                            break
                    n_degree[n] = len(data[n] & frontier) - cluster_degrees[c]
                p = max(frontier, key=lambda k: (n_degree[k], node_index[k]))
                if n_degree[p] > 0:
                    cp /= 2.0
            if (e_nk / density / (nn + 1)) < cp:
                break  # no good node found; cluster is finished

            # otherwise, add the node to the cluster
            cluster.add(p)
            nn, ne = (nn + 1), (ne + e_nk)

            cluster_degrees[p] = e_nk
            for n in data[p] & cluster:
                cluster_degrees[n] += 1

            frontier = set.union(*((data[n] - cluster) & unvisited for n in cluster))

        unvisited -= cluster
        res.append(list(cluster))

        num_clusters += 1
    return res


def __dpclus_w(g, d_threshold, cp_threshold, weights):
    res = []
    data = defaultdict(set)  # node id => neighboring node ids
    global_edges = defaultdict(dict)

    # to match the original DPClus output, we need to keep track of the order in
    # which nodes were read from the file, in order to maintain cluster order when
    # breaking ties in our priority ranking (values are decreasing, i.e. negative)
    node_index = dict()
    b_list = []  # the original algorithm indexes column b after column a... >:-(

    # read in graph
    for a, b, dt in g.edges(data=True):

        w = dt[weights]
        data[a].add(b)
        data[b].add(a)
        global_edges[a][b] = w
        global_edges[b][a] = w
        node_index.setdefault(a, -len(node_index))
        b_list.append(b)
    for b in b_list:
        node_index.setdefault(b, -len(node_index))

    unvisited = set(data)

    while unvisited:
        # get highest degree node
        # DIFFERENT: degree of node is sum of edge weights
        seed = max(
            unvisited,
            key=lambda k: (
                sum(global_edges[n][k] for n in data[k] & unvisited),
                node_index[k],
            ),
        )
        frontier = data[seed] & unvisited
        if not frontier:
            break  # no connections left to analyze

        edges, weights = defaultdict(lambda: defaultdict(int)), defaultdict(int)
        for a, b in combinations(unvisited, 2):
            if b not in data[a]:
                continue
            e = len(data[a] & data[b] & unvisited)
            edges[a][b], edges[b][a] = e, e
            weights[a] += e
            weights[b] += e

        max_w, _, node = max((w, node_index[n], n) for n, w in weights.items())
        if max_w > 0:
            seed = node
            frontier = data[seed] & unvisited

        cluster = {seed}
        cluster_degrees = {seed: 0}
        nn, ne = 1, 0.0  # number of nodes, edges in cluster (DIFFERENT: float)

        while frontier:
            # find higest priority node:
            # 0. number of edges between node and cluster nodes
            # 1. sum of edge weights between node and cluster nodes
            # 2. the node's index
            # 3. the node itself
            # DIFFERENT: e_nk = sum of edge weights
            e_nk, _, _, p = max(
                (
                    sum(global_edges[n][c] for c in data[n] & cluster),
                    sum(edges[n][c] for c in cluster),
                    node_index[n],
                    n,
                )
                for n in frontier
            )
            # e_nk,_,_,p = max((len(data[n] & cluster),
            # sum(edges[n][c] for c in cluster), node_index[n], n) for n in frontier)

            density = 2.0 * (ne + e_nk) / (nn * (nn + 1))
            if density < d_threshold:
                break  # adding the node gives too low density; cluster is finished

            # if all nodes only have one connecting edge in cluster, use "fine-tuning"
            # this orders by the number of neighbors in the frontier, minus the
            # connectedness of the attached (cluster) node within the cluster
            cp = cp_threshold
            if e_nk == 1 and len(cluster) > 1:
                n_degree = dict()  # node => fine-tuning parameter
                for n in frontier:
                    for c in cluster:  # find adjacent cluster node
                        if n in edges[c]:
                            break
                    n_degree[n] = len(data[n] & frontier) - cluster_degrees[c]
                p = max(frontier, key=lambda k: (n_degree[k], node_index[k]))
                if n_degree[p] > 0:
                    cp /= 2.0
            if (e_nk / density / (nn + 1)) < cp:
                break  # no good node found; cluster is finished

            # otherwise, add the node to the cluster
            cluster.add(p)
            nn, ne = (nn + 1), (ne + e_nk)

            cluster_degrees[p] = e_nk
            for n in data[p] & cluster:
                cluster_degrees[n] += 1

            frontier = set.union(*((data[n] - cluster) & unvisited for n in cluster))

        if len(cluster) > 1:
            # add overlapping nodes
            # frontier[2] stores our fine-tuning parameter, in this case
            frontier_nodes = set.union(*(data[c] for c in cluster)) - cluster
            # DIFFERENT: e_nk = sum of edge weights
            frontier = sorted(
                [
                    sum(global_edges[n][c] for c in data[n] & cluster),
                    sum(edges[n][c] for c in cluster),
                    0,
                    node_index[n],
                    n,
                ]
                for n in frontier_nodes
            )
            # "fine-tuning"
            fine_tuning = False
            if frontier and frontier[-1][0] == 1:
                for n in frontier:
                    for c in cluster:  # find adjacent cluster node
                        if n[4] in edges[c]:
                            break
                    n[2] = len(data[n[4]] & frontier_nodes) - cluster_degrees[c]
                frontier.sort(key=lambda k: (k[2], k[3]))
                fine_tuning = True
            # iterate through visited neighbors and update accordingly
            while frontier:
                e_nk, _, w, _, p = frontier.pop()
                cp = cp_threshold / 2.0 if fine_tuning and w > 0 else cp_threshold

                density = 2.0 * (ne + e_nk) / (nn * (nn + 1))
                if density < d_threshold or (e_nk / density / (nn + 1)) < cp:
                    continue

                # add node to the cluster
                cluster.add(p)
                nn, ne = (nn + 1), (ne + e_nk)

                cluster_degrees[p] = e_nk
                for n in data[p] & cluster:
                    cluster_degrees[n] += 1

                # update E_nk for the other nodes on the frontier
                for n in frontier:
                    if p in edges[n[4]]:
                        n[0] += 1

        unvisited -= cluster
        if len(cluster) > 1:
            res.append(list(cluster))

    return res



__authors__ = [
    "Marco Cardia <cardiamc@gmail.com>",
    "Francesco Sabiu <fsabiu@gmail.com>",
]


class weightedCommunity(object):
    def __init__(
        self,
        G,
        min_bel_degree,
        threshold_bel_degree,
        weightName="weight",
        save=False,
        outfile_name="weighted_communities.txt",
    ):
        """
        Constructor

        :param G: an igraph.Graph object
        :param min_bel_degree: the tolerance, in terms of beloging degree, required in order to add a node in a community
        :param threshold_bel_degree: the tolerance, in terms of beloging degree, required in order to add a node in a 'NLU' community
        :param weightName: Name of the attribute containing the weights
        """
        self.G = G
        self.N = len(self.G.vs)
        self.L = len(self.G.es)
        self.min_bel_degree = min_bel_degree
        self.threshold_bel_degree = threshold_bel_degree
        self.weightName = weightName
        self.save = save
        self.outfile_name = outfile_name
        self.communities = []
        self.strengths = []
        self.T = 0

        # Labels
        self.G.vs["label"] = ["F"] * self.N

    # Strength of a node
    def strength(self, node):
        strength = 0
        for j in self.G.neighbors(node):
            strength += self.G.es[self.G.get_eid(node, j)][self.weightName]
        return strength

    # Belonging degree
    def belonging_degree(self, node, community):
        strength = 0
        strength = sum(
            [
                self.G.es[self.G.get_eid(node, i)][self.weightName]
                for i in self.G.neighbors(node)
                if i in community
            ]
        )
        return strength / self.strength(node)

    # Modularity of a graph
    def modularity(self):
        import itertools
        import random

        Q_new = 0
        for c in self.communities:
            pair_components = list(itertools.combinations(c, 2))
            pair_components = random.sample(
                pair_components, 1000
            )  # Reducing execution time!
            for pair_c in pair_components:
                e_id = self.G.get_eid(pair_c[0], pair_c[1], error=False)
                weight = self.G.es[e_id][self.weightName] if e_id != -1 else 0
                Q_new += (
                    self.belonging_degree(pair_c[0], c)
                    * self.belonging_degree(pair_c[1], c)
                    * (
                        weight
                        - self.strength(pair_c[0])
                        * self.strength(pair_c[1])
                        / (2 * self.L)
                    )
                )

        Q_new = Q_new / (2 * self.L)

        return Q_new

    def allStrengths(self):
        for node in self.G.vs:
            self.strengths.append(self.strength(node))

        # Updating graph
        self.G.vs["strength"] = self.strengths

    def strongestNotLabeled(self):
        # Getting indices labeled with 'F'
        indices = [i for i in range(len(self.G.vs)) if self.G.vs[i]["label"] == "F"]

        # Getting strengths for such items
        ss = {}
        for i in indices:
            ss[self.strengths[i]] = i

        # Returning (one of) the strongest node index(es)
        return ss[max(ss.keys())]

    def nodesRemotion(self, c, min_bel_degree):
        old_c = 0
        while old_c != len(c):
            old_c = 0
            c_list = [el for el in c]
            for node in c_list:
                old_c = len(c_list)
                if self.belonging_degree(node, c) < min_bel_degree:
                    try:
                        c.remove(node)
                    except:
                        pass
        return c

    def initialCommunityDetection(self):
        c = set()
        # Getting strongest labeled with 'F'
        strongest = self.strongestNotLabeled()

        # Adding strongest to community
        c.add(strongest)
        self.G.vs[strongest]["label"] = "T"

        # Adding strongest's neighbors
        for neighbor in self.G.neighbors(strongest):
            c.add(neighbor)

        # Until convergence
        # if belonging degree of the community nodes is lower than min_bel_degree, we remove them.
        c = self.nodesRemotion(c, self.min_bel_degree)

        # If commuty is empty, add strongest to it
        if len(c) == 0:
            c.add(strongest)

        return c

    def find_initial_community_neighbors(self, c):
        ### Finding initial community neighbors
        c_neighbors = set()

        for node in c:
            for nb in self.G.neighbors(node):
                c_neighbors.add(nb)
        return c_neighbors

    def define_nu_nlu(self, c, c_neighbors):
        """### Find Nu (b > 0.5) and Nlu (0.4 < b < 0.5) starting from the set of neighbors"""
        nu = set()
        nlu = set()

        for nb in c_neighbors:
            # Nu set: neighbor with belonging degree >= min_bel_degree
            if self.belonging_degree(nb, c) >= self.min_bel_degree:
                nu.add(nb)
            # Nlu set neighbor with threshold < belonging degree < min_bel_degree
            elif self.belonging_degree(nb, c) > self.threshold_bel_degree:
                nlu.add(nb)

        return nu, nlu

    def add_nlu_to_community(self, c, nlu):
        while len(nlu) > 0:
            Q = self.modularity()

            candidate = nlu.pop()
            c.add(candidate)
            Q_new = self.modularity()
            if Q_new > Q:
                Q = Q_new
            else:
                c.remove(candidate)
        return c

    def expandCommunity(self, c):
        """##  Expanding the community"""
        c_old = 0
        nu, nlu = set(), set()

        while c_old != len(c):  # While it adds nodes belonging to nlu in the community
            while c_old != len(
                c
            ):  # While it adds nodes belonging to nu in the community
                c_neighbors = self.find_initial_community_neighbors(c)

                # Computing nu and nlu
                nu, nlu = self.define_nu_nlu(c, c_neighbors)

                # Add Nu nodes to initial community and return to neighbors search
                c_old = len(c)
                c = c.union(nu)

                # Asserting size
                if len(c) != len(c) + len(nu) - len(c.intersection(nu)):
                    print("Incompatible sizes")
                    return -1

            # Add the nodes of Nlu to the community only if their presence increases the community modularity Q0
            c_old = len(c)
            c = self.add_nlu_to_community(c, nlu)

        # Mark the community nodes with 'T' label
        self.G.vs["label"] = [
            v["label"] if not v.index in c else "T" for v in self.G.vs
        ]

        # Updating T label counter
        self.T = len([v for v in self.G.vs["label"] if v == "T"])

        return c

    def getCommunities(self):
        return self.communities

    def computeCommunity(self):
        if len(self.communities) == 0:  # Just the 1st time
            # Strenghts
            self.allStrengths()

        # Initial community detection
        c = self.initialCommunityDetection()

        # Community expansion
        c = self.expandCommunity(c)

        self.communities.append(c)

    def computeCommunities(self):
        while self.N != self.T:
            self.computeCommunity()

        # output communities
        if self.save and len(self.getCommunities()) > 0:
            out_file_com = open(self.outfile_name, "w")

            for cid, c in enumerate(self.getCommunities()):
                out_file_com.write("%d\t%s\n" % (cid, str(c)))

            out_file_com.flush()
            out_file_com.close()

        return self.getCommunities()



import numpy as np
import pandas as pd
from igraph import *


def condor_object(net):
    """Initialization of the condor object. The function gets a network in edgelist format encoded in a pandas
    dataframe. Returns a dictionary with an igraph network, names of the targets and regulators, list of edges,
    modularity, and vertex memberships.
    """

    # Error flags.
    assert (
        len(set(net.iloc[:, 0]).intersection(net.iloc[:, 1])) == 0
    ), "The network must be bipartite."
    assert not net.isnull().any().any(), "NaN values detected."
    assert not (
        "" in list(net.iloc[:, 0]) or "" in list(net.iloc[:, 1])
    ), "Empty strings detected."

    # Builds graph object.
    if net.shape[1] == 3:
        edges = list(zip(net.iloc[:, 0], net.iloc[:, 1], net.iloc[:, 2]))
        Gr = Graph.TupleList(edges, weights=True)
    else:
        edges = list(zip(net.iloc[:, 0], net.iloc[:, 1], [1 for _ in net.iloc[:, 1]]))
        Gr = Graph.TupleList(edges, weights=True)

    # Assigns color names (bipartite sets).
    reg_names = sorted(set(net.iloc[:, 1]))
    tar_names = sorted(set(net.iloc[:, 0]))
    Gr.vs["type"] = 0  # Tar
    for j in [i.index for i in Gr.vs if i["name"] in reg_names]:
        Gr.vs[j]["type"] = 1  # Reg

    index_dict = {k.index: k["name"] for k in Gr.vs}

    return {
        "G": Gr,
        "tar_names": tar_names,
        "reg_names": reg_names,
        "index_dict": index_dict,
        "edges": edges,
        "modularity": None,
        "reg_memb": None,
        "Qcoms": None,
    }


def bipartite_modularity(B, m, R, T, CO):
    """Computation of the bipartite modularity as described in ""Modularity and community detection in bipartite
    networks" by Michael J. Barber." """
    RtBT = R.transpose().dot(B.dot(T))
    Qcoms = (1 / m) * (np.diagonal(RtBT))
    Q = sum(Qcoms)
    Qcoms = Qcoms[Qcoms > 0]
    CO["Qcoms"] = Qcoms
    return Q, CO


def initial_community(CO, method="LCS"):
    """Computation of the initial community structure based on unipartite methods.
    The implementation using bipartite projection is not yet available, but project=False
    performs better modularity-wise (at least with the networks I worked on).
    """

    # Computes initial community assignement based on different methods. Default method is Louvain clustering
    # which is really fast. The others take several times more.
    vc = None
    if method == "LCS":
        vc = Graph.community_multilevel(CO["G"], weights="weight")
    if method == "LEC":
        vc = Graph.community_leading_eigenvector(CO["G"], weights="weight")
    if method == "FG":
        vc = Graph.community_fastgreedy(CO["G"], weights="weight").as_clustering()

    CO["modularity"] = vc.modularity

    # Stores initial community in the condor object.
    reg_index = [i.index for i in CO["G"].vs.select(type_in=[1])]
    reg_memb = [vc.membership[i] for i in reg_index]
    T0 = pd.DataFrame(zip(reg_index, reg_memb))
    T0.columns = ["index", "community"]
    CO["reg_memb"] = T0

    return CO


def brim(CO, deltaQmin="def", c=25):
    """Implementation of the BRIM algorithm to iteratively maximize bipartite modularity.
    Note that c is the maximum number of communities. Dynamic choice of c is not yet implemented.
    """

    # Gets modularity matrix, initial community matrix and index dictionary.
    B, m, T0, R0, gn, rg = matrices(CO, c)

    # Default deltaQmin.
    if deltaQmin == "def":
        deltaQmin = min(1 / len(CO["edges"]), 1e-5)

    # BRIM iterative process
    Qnow = 0
    deltaQ = 1
    p, q = B.shape
    R, T = None, None
    while deltaQ > deltaQmin:
        # Right sweep
        Tp = B.dot(T0)
        R = np.zeros((p, c))
        am = np.array(np.argmax(Tp, axis=1))
        for i in range(0, len(am)):
            R[i, am[i][0]] = 1
        # Left sweep
        Rp = B.transpose().dot(R)
        T = np.zeros((q, c))
        am = np.array(np.argmax(Rp, axis=1))
        for i in range(0, len(am)):
            T[i, am[i][0]] = 1
        T0 = T

        Qthen = Qnow
        Qnow, CO = bipartite_modularity(B, m, R, T, CO)
        deltaQ = Qnow - Qthen
        # print(Qnow)

    # Update modularity attribute
    CO["modularity"] = Qnow

    # Update membership dataframes.
    CO["tar_memb"] = pd.DataFrame(
        list(zip(list(gn), [R[i, :].argmax() for i in range(0, len(gn))]))
    )
    CO["reg_memb"] = pd.DataFrame(
        list(zip(list(rg), [T[i, :].argmax() for i in range(0, len(rg))]))
    )
    CO["tar_memb"].columns = ["tar", "com"]
    CO["reg_memb"].columns = ["reg", "com"]

    return CO


def matrices(CO, c):
    """Computation of modularity matrix and initial community matrix."""

    # Dimensions of the matrix
    p = len(CO["tar_names"])
    q = len(CO["reg_names"])

    # Index dictionaries for the matrix. Note that this set of indices is different of that in
    # the CO object (that one is for the igraph network.)
    rg = {CO["reg_names"][i]: i for i in range(0, q)}
    gn = {CO["tar_names"][i]: i for i in range(0, p)}

    # Computes bipartite adjacency matrix.
    A = np.matrix(np.zeros((p, q)))
    for edge in CO["edges"]:
        A[gn[edge[0]], rg[edge[1]]] = edge[2]

    # Computes bipartite modularity matrix.
    ki = A.sum(1)
    dj = A.sum(0)
    m = float(sum(ki))
    B = A - ((ki @ dj) / m)

    # Creates initial community T0 matrix.
    d = CO["index_dict"]
    if "index" in CO["reg_memb"].columns:
        ed = zip(
            [rg[j] for j in [d[i] for i in CO["reg_memb"].iloc[:, 0]]],
            CO["reg_memb"].iloc[:, 1],
        )
    else:
        ed = zip([rg[j] for j in CO["reg_memb"].iloc[:, 0]], CO["reg_memb"].iloc[:, 1])
    T0 = np.zeros((q, c))
    for edge in ed:
        T0[edge] = 1

    if "tar_memb" not in CO:
        # print("Matrices computed in", time.time() - t)
        return B, m, T0, 0, gn, rg
    ed = zip([gn[j] for j in CO["tar_memb"].iloc[:, 0]], CO["tar_memb"].iloc[:, 1])
    R0 = np.zeros((p, c))
    for edge in ed:
        R0[edge] = 1
    # print("Matrices computed in", time.time() - t)
    return B, m, T0, R0, gn, rg


def qscores(CO):
    """Computes the qscores (contribution of a vertex to its community modularity)
    for each vertex in the network."""

    B, m, T, R, gn, rg = matrices(CO, 6)
    CO["Qscores"] = {"reg_qscores": None, "tar_qscores": None}

    # Qscores for the regulators:
    Rq = R.transpose().dot(B) / (2 * m)
    Qj = list()
    for j, r in CO["reg_memb"].iterrows():
        Qjh = Rq[r["com"], j] / CO["Qcoms"][r["com"]]
        Qj.append(Qjh)
    CO["Qscores"]["reg_qscores"] = CO["reg_memb"].copy()
    CO["Qscores"]["reg_qscores"]["qscore"] = Qj

    # Qscores for the targets:
    Tq = B.dot(T) / (2 * m)
    Qi = list()
    for i, r in CO["tar_memb"].iterrows():
        Qih = Tq[i, r["com"]] / CO["Qcoms"][r["com"]]
        Qi.append(Qih)
    CO["Qscores"]["tar_qscores"] = CO["tar_memb"].copy()
    CO["Qscores"]["tar_qscores"]["qscore"] = Qi

    return CO



import queue
from operator import itemgetter
import networkx as nx
from collections import defaultdict
from multiprocessing import Pool

"""
source: https://github.com/pratham16/algorithms-detection-by-seed-expansion

**Overlapping Community Detection Using Seed Set Expansion (CIKM 2013)**
  - Joyce Jiyoung Whang, David F. Gleich, and Inderjit S. Dhillon
"""


def compute_local_pagerank(G, r, p, alpha, epsilon, max_push_count, q):
    for node, res in r.items():
        if res > epsilon * G.degree(node):
            q.put(node)

    push_count = 0
    while q.empty() == False and push_count < max_push_count:
        push_count += 1
        u = q.get()
        du = G.degree(u) * 1.0
        moving_prob = r[u] - 0.5 * epsilon * du
        r[u] = 0.5 * epsilon * du
        p[u] = p.get(u, 0.0) + (1 - alpha) * moving_prob

        neighbor_update = alpha * moving_prob / du

        for nbr in G.neighbors(u):
            nbrdeg = G.degree(nbr)
            rxold = r.get(nbr, 0.0)
            rxnew = rxold + neighbor_update
            r[nbr] = rxnew
        if rxnew > epsilon * nbrdeg >= rxold:
            q.put(nbr)

    return push_count


def cluster_from_sweep(G, p, community):
    p_sorted = sorted(p.items(), key=itemgetter(1), reverse=True)

    conductance = list()
    volume = list()
    cutsize = list()

    i = 0
    rank = dict()
    for node, ppr in p_sorted:
        rank[node] = i
        i += 1

    total_degree, curcutsize, curvolume, i = 2.0 * G.number_of_edges(), 0, 0, 0
    for node, ppr in p_sorted:
        deg = G.degree(node) * 1.0
        change = deg * 1.0
        for nbr in G.neighbors(node):
            if rank.get(nbr, -1) != -1 and rank.get(nbr, -1) < rank[node]:
                change -= 2

        curcutsize += change
        curvolume += deg

        cutsize.append(curcutsize)
        volume.append(curvolume)
        if curvolume == 0 or total_degree - curvolume == 0:
            conductance.append(1.0)
        else:
            conductance.append(curcutsize / min(curvolume, total_degree - curvolume))

    # print conductance
    lastind = len(conductance)
    mincond = float("INF")
    mincondind = 0
    for i in range(lastind):
        if conductance[i] <= mincond:
            mincond = conductance[i]
            mincondind = i

    for i in range(min(mincondind + 1, len(p_sorted))):
        community.append(p_sorted[i][0])

    return mincond


def pprgrow4(G, seed, alpha, targetvol):
    p = dict()
    r = dict()
    q = queue.Queue()
    community = list()

    for s in seed:
        r[s] = 1.0 / (len(seed) * 1.0)

    pr_eps = 1.0 / max(10.0 * targetvol, 100.0)
    maxsteps = 1.0 / (pr_eps * (1.0 - alpha))
    maxsteps = min(maxsteps, 0.5 * (2.0 ** 32 - 1.0))

    nsteps = compute_local_pagerank(G, r, p, alpha, pr_eps, int(maxsteps), q)
    if nsteps == 0:
        p = r

    # Scale the probabilities by their degree
    for node, pr in p.items():
        pr *= 1.0 / max(G.degree(node), 1.0)

    mincond = cluster_from_sweep(G, p, community)
    return community, mincond


def pprgrow(args):
    seed, G, stopping, nruns, alpha, maxexpand, fast = args
    expandseq = [2, 3, 4, 5, 10, 15]
    expands = list()
    curmod = 1
    while len(expands) < nruns:
        temp = [curmod * i for i in expandseq]
        for i in temp:
            expands.append(i)
        curmod *= 10

    expands = expands[:nruns]
    maxdeg = max(dict(G.degree(G.nodes())).values())
    bestcond = float("INF")

    if fast == True:
        expands = [1000]

    for ei in range(len(expands)):
        if fast == True:
            curexpand = expands[ei]
        else:
            if isinstance(seed, int):
                seed = [seed]
            curexpand = expands[ei] * len(seed) + maxdeg
        assert len(seed) > 0.0
        if curexpand > maxexpand:
            continue
        if stopping == "cond":
            curset, cond = pprgrow4(G, seed, alpha, curexpand)

            if cond < bestcond:
                bestcond = cond

    return curset


def growclusters(
    G, seeds, expansion, stopping, nworkers, nruns, alpha, maxexpand, fast
):
    if maxexpand == float("INF"):
        maxexpand = G.number_of_edges()

    ns = len(seeds)
    communities = list()

    if nworkers == 1:
        for i in range(ns):
            seed = seeds[i]
            curset = pprgrow((seed, G, stopping, nruns, alpha, maxexpand, fast))

            communities.append(curset)

    else:
        slen = len(seeds)
        args = zip(
            seeds,
            [G] * slen,
            [stopping] * slen,
            [nruns] * slen,
            [alpha] * slen,
            [maxexpand] * slen,
            [fast] * slen,
        )
        p = Pool(nworkers)
        if expansion == "ppr":
            communities = p.map(pprgrow, args)

    return communities


# TO DO: Write function to remove duplicate communities
def remove_duplicates(G, communities, delta):
    # Create node2com dictionary
    node2com = defaultdict(list)
    com_id = 0
    for comm in communities:
        for node in comm:
            node2com[node].append(com_id)
        com_id += 1

    deleted = dict()

    for i in range(len(communities)):
        comm = communities[i]
        if deleted.get(i, 0) == 0:
            nbrnodes = nx.node_boundary(G, comm)
            for nbr in nbrnodes:
                nbrcomids = node2com[nbr]
                for nbrcomid in nbrcomids:
                    if (
                        i != nbrcomid
                        and deleted.get(i, 0) == 0
                        and deleted.get(nbrcomid, 0) == 0
                    ):
                        nbrcom = communities[nbrcomid]
                        distance = 1.0 - (
                            len(set(comm) & set(nbrcom))
                            * 1.0
                            / (min(len(comm), len(nbrcom)) * 1.0)
                        )

                        if distance <= delta:
                            # Near duplicate communities found.
                            # Discard current algorithms
                            # Followed the idea of Lee et al. in GCE
                            deleted[i] = 1
                            for node in comm:
                                node2com[node].remove(i)
    for i in range(len(communities)):
        if deleted.get(i, 0) == 1:
            communities[i] = []

    communities = filter(lambda c: c != [], communities)  # Discard empty communities
    return communities


def neighbor_inflation(G, seeds):
    for i in range(len(seeds)):
        seed = seeds[i]
        egonet = [seed]
        for s in G.neighbors(seed):
            egonet.append(s)
        seeds[i] = list(set(egonet))

    return seeds
